import json
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Dict, Iterable, List, Optional, Tuple, Union
import numpy as np
from h3.api import numpy_int as h3

from timezonefinder.np_binary_helpers import (
    get_zone_ids_path,
    get_zone_positions_path,
    read_per_polygon_vector,
)
from timezonefinder.polygon_array import PolygonArray
from timezonefinder import utils, utils_clang
from timezonefinder.configs import (
    DEFAULT_DATA_DIR,
    SHORTCUT_H3_RES,
    CoordLists,
    CoordPairs,
    IntegerLike,
)

from timezonefinder.flatbuf.io.hybrid_shortcuts import (
    get_hybrid_shortcut_file_path,
    read_hybrid_shortcuts_binary,
)
from timezonefinder.zone_names import read_zone_names


class AbstractTimezoneFinder(ABC):
    # prevent dynamic attribute assignment (-> safe memory)
    """
    Abstract base class for TimezoneFinder instances
    """

    __slots__ = [
        "data_location",
        "shortcut_mapping",
        "in_memory",
        "_fromfile",
        "timezone_names",
        "zone_ids",
        "holes_dir",
        "boundaries_dir",
        "boundaries",
        "holes",
    ]

    zone_ids: np.ndarray
    shortcut_mapping: Dict[int, Union[int, np.ndarray]]
    """
    List of attribute names that store opened binary data files.
    """

    def __init__(
        self,
        bin_file_location: Optional[Union[str, Path]] = None,
        in_memory: bool = False,
    ):
        """
        Initialize the AbstractTimezoneFinder.
        :param bin_file_location: The path to the binary data files to use. If None, uses native package data.
        :param in_memory: ignored. All binary files will be read into memory (few MB). Only used for polygon coordinate data.
        """
        if bin_file_location is None:
            bin_file_location = DEFAULT_DATA_DIR
        self.data_location: Path = Path(bin_file_location)

        self.timezone_names = read_zone_names(self.data_location)

        # Load hybrid shortcut file - contains both zone IDs (for unique zones) and polygon arrays (for ambiguous zones)
        zone_ids_path = get_zone_ids_path(self.data_location)
        zone_ids_temp = read_per_polygon_vector(zone_ids_path)
        zone_id_dtype = zone_ids_temp.dtype

        path2shortcut = get_hybrid_shortcut_file_path(zone_id_dtype, self.data_location)
        self.shortcut_mapping = read_hybrid_shortcuts_binary(path2shortcut)

        zone_ids_path = get_zone_ids_path(self.data_location)
        self.zone_ids = read_per_polygon_vector(zone_ids_path)

    def _iter_boundary_ids_of_zone(self, zone_id: int) -> Iterable[int]:
        """
        Yield the boundary polygon IDs for a given zone ID.

        :param zone_id: ID of the zone
        :yield: boundary polygon IDs
        """
        # load only on demand. used when shortcuts contain zone IDs (hybrid optimization)
        zone_positions_path = get_zone_positions_path(self.data_location)
        zone_positions = np.load(zone_positions_path, mmap_mode="r")
        first_boundary_id_zone = zone_positions[zone_id]
        # read the id of the first boundary polygon of the consequent zone
        # NOTE: this has also been added for the last zone
        first_boundary_id_next = zone_positions[zone_id + 1]
        yield from range(first_boundary_id_zone, first_boundary_id_next)

    @property
    def nr_of_zones(self) -> int:
        """
        Get the number of timezones.

        :rtype: int
        """
        return len(self.timezone_names)

    @staticmethod
    def using_numba() -> bool:
        """
        Check if Numba is being used.

        :rtype: bool
        :return: True if Numba is being used to JIT compile helper functions
        """
        return utils.using_numba

    @staticmethod
    def using_clang_pip() -> bool:
        """
        :return: True if the compiled C implementation of the point in polygon algorithm is being used
        """
        return utils.inside_polygon == utils_clang.pt_in_poly_clang

    def zone_id_of(self, boundary_id: IntegerLike) -> int:
        """
        Get the zone ID of a polygon.

        :param boundary_id: The ID of the polygon.
        :type boundary_id: int
        :rtype: int
        """
        try:
            return int(self.zone_ids[boundary_id])
        except TypeError:
            raise ValueError(f"zone_ids is not set in directory {self.data_location}.")

    def zone_ids_of(self, boundary_ids: np.ndarray) -> np.ndarray:
        """
        Get the zone IDs of multiple boundary polygons.

        :param boundary_ids: An array of boundary polygon IDs.
        :return: array of corresponding timezone IDs.
        """
        return self.zone_ids[boundary_ids]

    def zone_name_from_id(self, zone_id: int) -> str:
        """
        Get the zone name from a zone ID.

        :param zone_id: The ID of the zone.
        :return: The name of the zone.
        :raises ValueError: If the timezone could not be found.
        """
        try:
            return self.timezone_names[zone_id]
        except IndexError:
            raise ValueError("timezone could not be found. index error.")

    def zone_name_from_boundary_id(self, boundary_id: IntegerLike) -> str:
        """
        Get the zone name from a boundary polygon ID.

        :param boundary_id: The ID of the boundary polygon.
        :return: The name of the zone.
        """
        zone_id = self.zone_id_of(boundary_id)
        return self.zone_name_from_id(zone_id)

    def _iter_boundaries_in_shortcut(self, *, lng: float, lat: float) -> Iterable[int]:
        """
        Iterate over boundary polygon IDs in the shortcut corresponding to the given coordinates.

        :param lng: The longitude of the point in degrees (-180.0 to 180.0).
        :param lat: The latitude of the point in degrees (90.0 to -90.0).
        :yield: Boundary polygon IDs.
        """
        hex_id = h3.latlng_to_cell(lat, lng, SHORTCUT_H3_RES)

        # Handle shortcuts (hybrid structure) - if it's a zone ID, get all polygons for that zone
        shortcut_value = self.shortcut_mapping.get(hex_id)
        if shortcut_value is None:
            return
        elif isinstance(shortcut_value, int):
            # Zone ID - get all boundary polygons for this zone
            # Most polygons will be quickly ruled out by bbox check
            yield from self._iter_boundary_ids_of_zone(shortcut_value)
        else:
            # Polygon array
            yield from shortcut_value

    @abstractmethod
    def timezone_at(self, *, lng: float, lat: float) -> Optional[str]:
        """looks up in which timezone the given coordinate is included in

        :param lng: longitude of the point in degree (-180.0 to 180.0)
        :param lat: latitude in degree (90.0 to -90.0)
        :return: the timezone name of a matching polygon or None
        """
        ...

    def timezone_at_land(self, *, lng: float, lat: float) -> Optional[str]:
        """computes in which land timezone a point is included in

        Especially for large polygons it is expensive to check if a point is really included.
        To speed things up there are "shortcuts" being used (stored in a binary file),
        which have been precomputed and store which timezone polygons have to be checked.

        :param lng: longitude of the point in degree (-180.0 to 180.0)
        :param lat: latitude in degree (90.0 to -90.0)
        :return: the timezone name of a matching polygon or
            ``None`` when an ocean timezone ("Etc/GMT+-XX") has been matched.
        """
        tz_name = self.timezone_at(lng=lng, lat=lat)
        if tz_name is not None and utils.is_ocean_timezone(tz_name):
            return None
        return tz_name

    def unique_timezone_at(self, *, lng: float, lat: float) -> Optional[str]:
        """returns the name of a unique zone within the corresponding shortcut

        :param lng: longitude of the point in degree (-180.0 to 180.0)
        :param lat: latitude in degree (90.0 to -90.0)
        :return: the timezone name of the unique zone or ``None`` if there are no or multiple zones in this shortcut
        """
        lng, lat = utils.validate_coordinates(lng, lat)
        hex_id = h3.latlng_to_cell(lat, lng, SHORTCUT_H3_RES)

        # Shortcuts behavior (hybrid structure with precomputed uniqueness)
        shortcut_value = self.shortcut_mapping.get(hex_id)
        if shortcut_value is None:
            return None
        elif isinstance(shortcut_value, int):
            # Zone ID - this is a precomputed unique zone
            unique_id = shortcut_value
        else:
            # Polygon array - by definition not unique (would be stored as int if unique)
            return None

        return self.zone_name_from_id(unique_id)

    def cleanup(self) -> None:
        """Clean up resources. Override in subclasses as needed."""
        pass

    def __enter__(self):
        """Enter the runtime context for the TimezoneFinder."""
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Exit the runtime context and clean up resources."""
        self.cleanup()
        return False


class TimezoneFinderL(AbstractTimezoneFinder):
    """a 'light' version of the TimezoneFinder class for quickly suggesting a timezone for a point on earth

    Instead of using timezone polygon data like ``TimezoneFinder``,
    this class only uses a precomputed 'shortcut' to suggest a probable result:
    the most common zone in a rectangle of a half degree of latitude and one degree of longitude
    """

    def __init__(
        self, bin_file_location: Optional[str] = None, in_memory: bool = False
    ):
        super().__init__(bin_file_location, in_memory)

    def timezone_at(self, *, lng: float, lat: float) -> Optional[str]:
        """instantly returns the name of the most common zone within the corresponding shortcut

        Note: 'most common' in this context means that the boundary polygons with the most coordinates in sum
            occurring in the corresponding shortcut belong to this zone.

        :param lng: longitude of the point in degree (-180.0 to 180.0)
        :param lat: latitude in degree (90.0 to -90.0)
        :return: the timezone name of the most common zone or None if there are no timezone polygons in this shortcut
        """
        lng, lat = utils.validate_coordinates(lng, lat)
        # Inline fast-path to minimize helper overhead
        hex_id = h3.latlng_to_cell(lat, lng, SHORTCUT_H3_RES)

        shortcut_value = self.shortcut_mapping.get(hex_id)
        if shortcut_value is None:
            return None
        elif isinstance(shortcut_value, int):
            # Zone ID - unique zone case
            return self.zone_name_from_id(shortcut_value)
        else:
            # Polygon array - get the last polygon (most common zone)
            if len(shortcut_value) == 0:
                return None
            poly_of_biggest_zone = shortcut_value[-1]
            # poly_of_biggest_zone is a numpy scalar from array indexing, but mypy sees it as ndarray
            # This is safe: array element access returns a numpy integer scalar compatible with IntegerLike
            most_common_id = self.zone_id_of(poly_of_biggest_zone)  # type: ignore[arg-type]
            return self.zone_name_from_id(most_common_id)


class TimezoneFinder(AbstractTimezoneFinder):
    """Class for quickly finding the timezone of a point on earth offline.

    Because of indexing ("shortcuts"), not all timezone polygons have to be tested during a query.

    Opens the required timezone polygon data in binary files to enable fast access.
    For a detailed documentation of data management please refer to the code documentation of
    `file_converter.py <https://github.com/jannikmi/timezonefinder/blob/master/scripts/file_converter.py>`__

    :ivar binary_data_attributes: the names of all attributes which store the opened binary data files

    :param bin_file_location: path to the binary data files to use, None if native package data should be used
    :param in_memory: Whether to completely read and keep the coordinate data in memory as numpy arrays.
    """

    # __slots__ declared in parents are available in child classes. However, child subclasses will get a __dict__
    # and __weakref__ unless they also define __slots__ (which should only contain names of any additional slots).
    __slots__ = [
        "hole_registry",
        "_boundaries_file",
        "_holes_file",
    ]

    def __init__(
        self, bin_file_location: Optional[str] = None, in_memory: bool = False
    ):
        super().__init__(bin_file_location, in_memory)
        self.holes_dir = utils.get_holes_dir(self.data_location)
        self.boundaries_dir = utils.get_boundaries_dir(self.data_location)
        self.boundaries = PolygonArray(
            data_location=self.boundaries_dir, in_memory=in_memory
        )
        self.holes = PolygonArray(data_location=self.holes_dir, in_memory=in_memory)

        # stores for which polygons (how many) holes exits and the id of the first of those holes
        # since there are very few entries it is feasible to keep them in the memory
        self.hole_registry = self._load_hole_registry()

    def __del__(self) -> None:
        """Clean up resources when the object is destroyed."""
        del self.boundaries
        del self.holes
        del self.hole_registry

    def _load_hole_registry(self) -> Dict[int, Tuple[int, int]]:
        """
        Load and convert the hole registry from JSON file, converting keys to int.
        """
        path = utils.get_hole_registry_path(self.data_location)
        with open(path, encoding="utf-8") as json_file:
            hole_registry_tmp = json.loads(json_file.read())
        # convert the json string keys to int
        return {int(k): v for k, v in hole_registry_tmp.items()}

    @property
    def nr_of_polygons(self) -> int:
        return len(self.boundaries)

    @property
    def nr_of_holes(self) -> int:
        return len(self.holes)

    def coords_of(self, boundary_id: IntegerLike = 0) -> np.ndarray:
        """
        Get the coordinates of a boundary polygon from the FlatBuffers collection.

        :param boundary_id: The index of the polygon.
        :return: Array of coordinates.
        """
        return self.boundaries.coords_of(boundary_id)

    def _iter_hole_ids_of(self, boundary_id: IntegerLike) -> Iterable[int]:
        """
        Yield the hole IDs for a given boundary polygon id.

        :param boundary_id: id of the boundary polygon
        :yield: Hole IDs
        """
        try:
            amount_of_holes, first_hole_id = self.hole_registry[int(boundary_id)]
        except KeyError:
            return
        for i in range(amount_of_holes):
            yield first_hole_id + i

    def _holes_of_poly(self, boundary_id: IntegerLike) -> Iterable[np.ndarray]:
        """
        Get the hole coordinates of a boundary polygon from the FlatBuffers collection.

        :param boundary_id: id of the boundary polygon
        :yield: Generator of hole coordinates
        """
        for hole_id in self._iter_hole_ids_of(boundary_id):
            yield self.holes.coords_of(hole_id)

    def get_polygon(
        self, boundary_id: IntegerLike, coords_as_pairs: bool = False
    ) -> List[Union[CoordPairs, CoordLists]]:
        """
        Get the polygon coordinates of a given boundary polygon including its holes.

        :param boundary_id:  ID of the boundary polygon
        :param coords_as_pairs: If True, returns coordinates as pairs (lng, lat).
            If False, returns coordinates as separate lists of longitudes and latitudes.
        :return: List of polygon coordinates
        """
        list_of_converted_polygons = []
        if coords_as_pairs:
            conversion_method = utils.convert2coord_pairs
        else:
            conversion_method = utils.convert2coords
        list_of_converted_polygons.append(
            conversion_method(self.coords_of(boundary_id=boundary_id))
        )

        for hole in self._holes_of_poly(boundary_id):
            list_of_converted_polygons.append(conversion_method(hole))

        return list_of_converted_polygons

    def get_geometry(
        self,
        tz_name: Optional[str] = "",
        tz_id: Optional[int] = 0,
        use_id: bool = False,
        coords_as_pairs: bool = False,
    ) -> List[List[Union[CoordPairs, CoordLists]]]:
        """retrieves the geometry of a timezone: multiple boundary polygons with holes

        :param tz_name: one of the names in ``timezone_names.json`` or ``self.timezone_names``
        :param tz_id: the id of the timezone (=index in ``self.timezone_names``)
        :param use_id: if ``True`` uses ``tz_id`` instead of ``tz_name``
        :param coords_as_pairs: determines the structure of the polygon representation
        :return: a data structure representing the multipolygon of this timezone
            output format: ``[ [polygon1, hole1, hole2...], [polygon2, ...], ...]``
            and each polygon and hole is itself formatted like: ``([longitudes], [latitudes])``
            or ``[(lng1,lat1), (lng2,lat2),...]`` if ``coords_as_pairs=True``.
        """

        if use_id:
            if not isinstance(tz_id, int):
                raise TypeError("the zone id must be given as int.")
            if tz_id < 0 or tz_id >= self.nr_of_zones:
                raise ValueError(
                    f"the given zone id {tz_id} is invalid (value range: 0 - {self.nr_of_zones - 1}."
                )
        else:
            if tz_name is None:
                raise ValueError("no timezone name given.")
            try:
                tz_id = self.timezone_names.index(tz_name)
            except ValueError:
                raise ValueError("The timezone '", tz_name, "' does not exist.")
        if tz_id is None:
            raise ValueError("no timezone id given.")

        return [
            self.get_polygon(boundary_id, coords_as_pairs)
            for boundary_id in self._iter_boundary_ids_of_zone(tz_id)
        ]

    def inside_of_polygon(self, boundary_id: IntegerLike, x: int, y: int) -> bool:
        """
        Check if a point is inside a boundary polygon.

        :param boundary_id: boundary polygon ID
        :param x: X-coordinate of the point
        :param y: Y-coordinate of the point
        :return: True if the point lies inside the boundary polygon, False if outside or in a hole.
        """
        # avoid running the expensive PIP algorithm at any cost
        # -> check bboxes first
        if self.boundaries.outside_bbox(boundary_id, x, y):
            return False

        # NOTE: holes are much smaller (fewer points) -> less expensive to check
        # -> check holes before the boundary
        hole_id_iter = self._iter_hole_ids_of(boundary_id)
        if self.holes.in_any_polygon(hole_id_iter, x, y):
            # the point is within one of the holes
            # it is excluded fromn this boundary polygon
            return False

        return self.boundaries.pip(boundary_id, x, y)

    def timezone_at(self, *, lng: float, lat: float) -> Optional[str]:
        """
        Find the timezone for a given point using hybrid shortcuts, considering both land and ocean timezones.

        Uses precomputed hybrid shortcuts to reduce the number of polygons checked. Returns the timezone name
        of the matched polygon, which may be an ocean timezone ("Etc/GMT+-XX") if applicable.

        Since ocean timezones span the whole globe, some timezone will always be matched!
        `None` can only be returned when using custom timezone data without such ocean timezones.

        :param lng: longitude of the point in degrees (-180.0 to 180.0)
        :param lat: latitude of the point in degrees (90.0 to -90.0)
        :return: the timezone name of the matched polygon, or None if no match is found.
        """
        # NOTE: performance critical code. avoid helper function call overhead as much as possible
        lng, lat = utils.validate_coordinates(lng, lat)
        hex_id = h3.latlng_to_cell(lat, lng, SHORTCUT_H3_RES)

        # Get shortcut value (hybrid optimization)
        shortcut_value = self.shortcut_mapping.get(hex_id)
        if shortcut_value is None:
            # NOTE: hypothetical case, with ocean data every shortcut maps to at least one boundary polygon
            return None

        if isinstance(shortcut_value, int):
            # Direct zone ID - optimal case for performance
            return self.zone_name_from_id(shortcut_value)

        # Polygon array case - need to check polygons
        possible_boundaries = shortcut_value
        nr_possible_polygons = len(possible_boundaries)
        if nr_possible_polygons == 0:
            return None
        # NOTE: the length 1 case can never occur here, since this is covered by the unique zone shortcut

        # create a list of all the timezone ids of all possible boundary polygons
        zone_ids = self.zone_ids_of(possible_boundaries)

        last_zone_change_idx = utils.get_last_change_idx(zone_ids)
        # NOTE: the case last_zone_change_idx == 0 is covered by the unique zone shortcut

        # ATTENTION: the polygons are stored converted to 32-bit ints,
        # convert the query coordinates in the same fashion in order to make the data formats match
        # x = longitude  y = latitude  both converted to 8byte int
        x = utils.coord2int(lng)
        y = utils.coord2int(lat)

        # check until the point is included in one of the possible boundary polygons
        for i, boundary_id in enumerate(possible_boundaries):
            if i >= last_zone_change_idx:
                # avoid expensive PIP checks when no other zone can be matched anymore
                break

            if self.inside_of_polygon(boundary_id, x, y):
                zone_id = zone_ids[i]
                return self.zone_name_from_id(int(zone_id))

        # since it is the last possible option,
        # the polygons of the last possible zone don't actually have to be checked
        # -> instantly return the last zone
        zone_id = zone_ids[-1]
        return self.zone_name_from_id(int(zone_id))

    def certain_timezone_at(self, *, lng: float, lat: float) -> Optional[str]:
        """checks in which timezone polygon the point is certainly included in using hybrid shortcuts

        .. note:: this is only meaningful when you have compiled your own timezone data
            where there are areas without timezone polygon coverage.
            Otherwise, some timezone will always be matched and the functionality is equal to using `.timezone_at()`
            -> useless to actually test all polygons.

        .. note:: using this function is less performant than `.timezone_at()`

        :param lng: longitude of the point in degree
        :param lat: latitude of the point in degree
        :return: the timezone name of the polygon the point is included in or `None`
        """
        lng, lat = utils.validate_coordinates(lng, lat)
        hex_id = h3.latlng_to_cell(lat, lng, SHORTCUT_H3_RES)

        # Get shortcut value (hybrid optimization)
        shortcut_value = self.shortcut_mapping.get(hex_id)
        if shortcut_value is None:
            return None

        # ATTENTION: the polygons are stored converted to 32-bit ints,
        # convert the query coordinates in the same fashion in order to make the data formats match
        # x = longitude  y = latitude  both converted to 8byte int
        x = utils.coord2int(lng)
        y = utils.coord2int(lat)

        # check if the query point is found to be truly included in one of the possible boundary polygons
        if isinstance(shortcut_value, int):
            # For zone IDs, iterate directly over boundary polygons for that zone
            # Most polygons will be quickly ruled out by bbox check
            boundary_ids = self._iter_boundary_ids_of_zone(shortcut_value)
        else:
            # Polygon array case - iterate directly over the array
            boundary_ids = shortcut_value

        for boundary_id in boundary_ids:
            if self.inside_of_polygon(boundary_id, x, y):
                zone_id = self.zone_id_of(boundary_id)
                return self.zone_name_from_id(zone_id)

        # none of the boundary polygon candidates truly matched
        return None



import argparse
import contextlib
import os
import sys
import tempfile
from typing import Callable, Generator

from timezonefinder import (
    TimezoneFinderL,
    timezone_at,
    certain_timezone_at,
    timezone_at_land,
)


@contextlib.contextmanager
def redirect_stdout_to_temp_file() -> Generator[str, None, None]:
    """
    Context manager that redirects stdout to a temporary file for the duration of the context.
    The temporary file is created but not deleted when the context exits.
    Returns the path to the temporary file.
    """
    # Save the original stdout
    original_stdout = sys.stdout

    # Create a temporary file that will NOT be automatically deleted
    temp_fd, temp_path = tempfile.mkstemp(text=True)
    temp_file = os.fdopen(temp_fd, "w")

    try:
        # Redirect stdout to the temporary file
        sys.stdout = temp_file
        yield temp_path
    finally:
        # Restore the original stdout and close the file
        sys.stdout = original_stdout
        temp_file.close()


def get_timezone_function(function_id: int) -> Callable:
    """
    Get the appropriate timezone function based on the function ID.
    Uses global functions when available, otherwise creates instances as needed.
    """
    # Use global functions for TimezoneFinder methods
    if function_id == 0:
        return timezone_at
    elif function_id == 1:
        return certain_timezone_at
    elif function_id == 5:
        return timezone_at_land

    # For TimezoneFinderL methods, still create an instance
    tf_instance = TimezoneFinderL()
    functions = {
        3: tf_instance.timezone_at,
        4: tf_instance.timezone_at_land,
    }
    return functions[function_id]


def main() -> None:
    parser = argparse.ArgumentParser(description="parse TimezoneFinder parameters")
    parser.add_argument("lng", type=float, help="longitude to be queried")
    parser.add_argument("lat", type=float, help="latitude to be queried")
    parser.add_argument("-v", action="store_true", help="verbosity flag")
    parser.add_argument(
        "-f",
        "--function",
        type=int,
        choices=[0, 1, 3, 4, 5],
        default=0,
        help="function to be called:"
        "0: TimezoneFinder.timezone_at(), "
        "1: TimezoneFinder.certain_timezone_at(), "
        "2: removed, "
        "3: TimezoneFinderL.timezone_at(), "
        "4: TimezoneFinderL.timezone_at_land(), "
        "5: TimezoneFinder.timezone_at_land(), ",
    )
    parsed_args = parser.parse_args()  # takes input from sys.argv
    timezone_function = get_timezone_function(parsed_args.function)

    verbose_mode = parsed_args.v

    # Always redirect stdout to a temp file
    with redirect_stdout_to_temp_file() as temp_file_path:
        print("\n" + "=" * 60)
        print("TIMEZONEFINDER LOOKUP DETAILS")
        print("-" * 60)
        print(f"Coordinates: {parsed_args.lat:.6f}°, {parsed_args.lng:.6f}° (lat, lng)")
        print(
            f"Function {timezone_function.__name__} (function ID: {parsed_args.function})"
        )

        # Execute the timezone function
        tz = timezone_function(lng=parsed_args.lng, lat=parsed_args.lat)

        if tz:
            print(f"Result: Found timezone '{tz}'")
        else:
            print("Result: No timezone found at this location")
        print("=" * 60)

    if verbose_mode:
        # In verbose mode, print the contents of the temp file
        try:
            with open(temp_file_path) as f:
                captured_output = f.read().strip()
                if captured_output:
                    print(captured_output)
        except Exception as e:
            print(f"Warning: Could not read captured output: {e}")
    else:
        # In non-verbose mode, just print the result
        print(tz if tz else "")

    # Always clean up the temp file
    try:
        os.remove(temp_file_path)
    except Exception:
        pass



"""Tests for verifying the contents of the package distribution.

This module tests that:
1. All required files are included in both source (sdist) and binary (wheel) distributions
2. No unwanted files are included (cache, .env, temporary files, etc.) in either distribution type

The module uses parameterized tests and global constants to minimize code
duplication and make the tests more maintainable. It builds both distribution types
and verifies their contents independently.
"""

from pathlib import Path
import tarfile
import tempfile
import zipfile
from typing import Iterator, List, Set

import pytest
from tests.auxiliaries import (
    PROJECT_ROOT,
    any_filter_paths,
    build_sdist,
    build_wheel,
    file_path_iterator,
    filter_paths,
    matches_pattern,
)

# Mark all tests in this module as integration tests
pytestmark = pytest.mark.integration

GITIGNORE_PATH = PROJECT_ROOT / ".gitignore"


SDIST_TYPE = "sdist"
WHEEL_TYPE = "wheel"
DIST_TYPES = [SDIST_TYPE, WHEEL_TYPE]


#######################
# FILE PATTERN CONSTANTS
#######################

# Additional patterns to ignore beyond what's in .gitignore
# TODO test
# no files matching these patterns should be included in the distribution
UNWANTED_DIST_PATTERNS = {
    ".github",
    ".git",
    ".vscode/",
    "build/",
    "examples/",
    "docs/",
    "scripts/",
    "prototypes/",
    # "tests/", # NOTE: tests should be included in the package for users to validate the package
    ".coveragerc",
    ".editorconfig",
    ".gitignore",
    ".pre-commit-config.yaml",
    "CHANGELOG*",
    "CONTRIBUTING.*",
    "Agents.*",
    "Makefile",
    "parse_data.sh",
    "readthedocs.yaml",
    "test_musllinux_wheel.sh",
    "tox.ini",
    "uv.lock",
}

ALLOWED_IGNORED_PATTERNS = {
    "*.egg-info/",
    "*.so",  # Compiled shared objects, are ok in wheels
}  # not under version control, but should be included in the distribution


#######################
# ESSENTIAL PATTERNS
#######################

# TODO test
# these files are not part of the source code, but must be included in the distribution
EXPECTED_DIST_PATTERNS = {
    "PKG-INFO",
} | ALLOWED_IGNORED_PATTERNS


# all files matching these patterns should be included in source distribution
ESSENTIAL_SOURCE_PATTERNS = {
    "setup.py",
    "pyproject.toml",
    # "MANIFEST.in",
    # "requirements.txt",
    "*py.typed",
    "README*",
    "*LICENSE*",
    # FILE EXTENSIONS:
    # source:
    "*.py",  # all Python source files
    "*.c",  # C source files
    "*.h",  # C header files
    "*.so",  # Compiled shared objects
    # data files:
    "*.npy",  # Numpy binary data files
    "*.fbs",  # Flatbuffer schema files
    # FIXME: does not catch missing hole_registry.json
    "*.json",  # used for hole registry
    "*.txt",  # Text files (used for timezone names)
}

# these files are not included in wheels
WHEEL_EXCEPTION_PATTERNS = {
    "pyproject.toml",
    "setup.py",
    "README*",
    # wheels define what is installed. tests should not be included
    "tests/*",
}


def load_gitignore_patterns() -> Set[str]:
    """
    Load patterns from a .gitignore file.

    Args:
        gitignore_path: Path to the .gitignore file (default: PROJECT_ROOT/.gitignore)

    Returns:
        A set of patterns loaded from the .gitignore file.
    """
    with open(GITIGNORE_PATH, encoding="utf-8") as f:
        # Read lines and strip whitespace
        return {line.strip() for line in f if line.strip() and not line.startswith("#")}


# any file not under version control should not be included in the distribution
NON_VERSION_CONTROL_PATTERNS = load_gitignore_patterns()
IGNORED_PATTERNS = UNWANTED_DIST_PATTERNS | NON_VERSION_CONTROL_PATTERNS
# NOTE: some patterns are not under version control, but should be included in the distribution
UNWANTED_DIST_PATTERNS_FINAL = IGNORED_PATTERNS - ALLOWED_IGNORED_PATTERNS


def filter_ignore_patterns(paths: Iterator[Path]) -> Iterator[Path]:
    """
    Filter out paths that match any pattern in IGNORE_PATTERNS.

    Args:
        paths: An iterator of Path objects to filter

    Yields:
        Path objects that don't match any ignore pattern

    Examples:
        # Get all files except those matching ignore patterns
        all_files = iterate_files_by_pattern()
        valid_files = list(filter_ignore_patterns(all_files))
    """
    return any_filter_paths(paths, IGNORED_PATTERNS, include_matches=False)


def get_distributable_files() -> Iterator[Path]:
    """
    Get all files that should be included in the distribution.

    This function filters out files matching IGNORE_PATTERNS.

    Returns:
        Iterator of Path objects for files that should be included in the distribution
    """
    all_files = file_path_iterator(PROJECT_ROOT, relative=True)

    # Filter out ignored files
    return filter_ignore_patterns(all_files)


def iter_expected_distribution_files() -> Iterator[Path]:
    """
    Get all essential source files that should be included in the distribution.

    This function filters out files matching IGNORE_PATTERNS.

    Returns:
        Iterator of Path objects for essential source files
    """
    all_files = get_distributable_files()

    # Filter out ignored files
    return any_filter_paths(all_files, ESSENTIAL_SOURCE_PATTERNS, include_matches=True)


def extract_archive(archive_path: Path) -> List[Path]:
    """Extract the tar.gz archive in the given path and return a list of the contained files."""
    with tarfile.open(archive_path, "r:gz") as tar:
        # Get the name of the top-level directory in the archive
        top_level_dirs = {member.name.split("/")[0] for member in tar.getmembers()}
        if len(top_level_dirs) == 0:
            raise ValueError("The archive does not contain any files.")

        # work in a temporary directory
        with tempfile.TemporaryDirectory() as tmpdir:
            extract_to = Path(tmpdir)
            # Extract all files with the 'data' filter to allow all files but avoid the deprecation warning
            tar.extractall(path=extract_to, filter="data")

            # Find the package directory (it should contain setup.py or pyproject.toml)
            pkg_dir = None
            for dir_name in top_level_dirs:
                pkg_dir = Path(extract_to) / dir_name
                if (pkg_dir / "pyproject.toml").exists():
                    break

            if pkg_dir is None:
                raise ValueError("No package directory found in the archive.")

            archive_files = file_path_iterator(pkg_dir, relative=True)
            file_list = list(archive_files)

    assert len(file_list) > 0, "The archive does not contain any files."
    return file_list


def extract_wheel(wheel_path: Path) -> List[Path]:
    """Extract the wheel (.whl) file in the given path and return a list of the contained files."""
    with zipfile.ZipFile(wheel_path) as wheel:
        # List all files in the wheel
        wheel_files = wheel.namelist()

        if len(wheel_files) == 0:
            raise ValueError("The wheel does not contain any files.")

        # work in a temporary directory
        with tempfile.TemporaryDirectory() as tmpdir:
            extract_to = Path(tmpdir)
            # Extract all files
            wheel.extractall(path=extract_to)
            # wheel.extractall(path=PROJECT_ROOT/"tmp"/"wheel")

            # Get list of all files (relative paths)
            archive_files = file_path_iterator(extract_to, relative=True)
            file_list = list(archive_files)

    assert len(file_list) > 0, "The wheel does not contain any files."
    return file_list


class DistributionFilesFixture:
    """A fixture class to manage the distribution files testing context.

    This singleton class builds the distribution once and provides access
    to the files for all tests, improving performance significantly.
    """

    def __init__(self, dist_type="sdist"):
        """Initialize the fixture with empty attributes."""
        self.dist_type = dist_type  # "sdist" or "wheel"
        self.temp_dir = None
        self.archive_path = None
        self.extract_dir = None
        self.project_files = None
        self.archive_files = None
        self.package_name = None
        self._initialized = False

    def initialize(self):
        """Initialize the fixture by building and extracting the distribution."""
        if self._initialized:
            return

        # Build the distribution based on type
        if self.dist_type == "sdist":
            self.archive_path = build_sdist()
            self.archive_files = extract_archive(self.archive_path)
        elif self.dist_type == "wheel":
            self.archive_path = build_wheel()
            self.archive_files = extract_wheel(self.archive_path)
        else:
            raise ValueError(f"Unknown distribution type: {self.dist_type}")

        self._initialized = True
        print(
            f"Built and extracted {self.dist_type} distribution with {len(self.archive_files)} files"
        )


# Create singleton instances for the fixtures
sdist_fixture = DistributionFilesFixture("sdist")
wheel_fixture = DistributionFilesFixture("wheel")

fixtures = {SDIST_TYPE: sdist_fixture, WHEEL_TYPE: wheel_fixture}

# parameterised pytest test case for testing that all distribution files do not match any of the unwanted patterns


@pytest.mark.parametrize("pattern", UNWANTED_DIST_PATTERNS_FINAL)
@pytest.mark.parametrize("dist_type", DIST_TYPES)
def test_no_unwanted_files_in_distribution(
    pattern: str,
    dist_type: str,
):
    """Test that no unwanted files are included in the distribution."""

    # Get all files in the distribution
    fixture = fixtures[dist_type]
    fixture.initialize()
    dist_files = fixture.archive_files

    # Filter out files that match the ignore patterns
    ignored_files = filter_paths(dist_files, pattern, include_matches=True)

    ignored_file_repr = [str(f) for f in ignored_files]
    nr_ignored_files = len(ignored_file_repr)
    assert nr_ignored_files == 0, (
        f"Found {nr_ignored_files} unwanted files matching pattern '{pattern}' in {dist_type}: {', '.join(ignored_file_repr)}"
    )


# parameterised pytest test case for testing that all essential source files are included in the distribution
@pytest.mark.parametrize("expected_file", iter_expected_distribution_files())
@pytest.mark.parametrize("dist_type", DIST_TYPES)
def test_essential_files_in_distribution(expected_file: Path, dist_type: str):
    """Test that all essential source files are included in the distribution."""
    pattern = str(expected_file)
    # Get all files in the distribution
    fixture = fixtures[dist_type]
    fixture.initialize()
    dist_files = fixture.archive_files

    # NOTE: in wheels the files may be in a subdirectory, relax the pattern matching
    if dist_type == WHEEL_TYPE:
        if any(matches_pattern(expected_file, p) for p in WHEEL_EXCEPTION_PATTERNS):
            # some files are ok to not exist in wheels
            print(f"Skipping {expected_file} in {dist_type} due to wheel exceptions")
            return
        patterns = [
            pattern,
            f"**/{pattern}",
        ]
    else:
        patterns = [pattern]
    matched_files = any_filter_paths(dist_files, patterns, include_matches=True)
    matched_file_repr = [str(f) for f in matched_files]
    nr_matched_files = len(matched_file_repr)
    assert nr_matched_files < 2, (
        f"multiple files matched pattern '{pattern}' in {dist_type}: {', '.join(matched_file_repr)}"
    )
    assert nr_matched_files == 1, (
        f"Essential file '{pattern}' not found in {dist_type}."
    )



"""Utilities for reading and writing FlatBuffer assets."""

from .polygons import (
    flatten_polygon_coords,
    reshape_to_polygon_coords,
    get_coordinate_path,
    write_polygon_collection_flatbuffer,
    get_polygon_collection,
    read_polygon_array_from_binary,
)
from .hybrid_shortcuts import (
    get_hybrid_shortcut_file_path,
    write_hybrid_shortcuts_flatbuffers,
    read_hybrid_shortcuts_binary,
)

__all__ = [
    "flatten_polygon_coords",
    "reshape_to_polygon_coords",
    "get_coordinate_path",
    "write_polygon_collection_flatbuffer",
    "get_polygon_collection",
    "read_polygon_array_from_binary",
    "get_hybrid_shortcut_file_path",
    "write_hybrid_shortcuts_flatbuffers",
    "read_hybrid_shortcuts_binary",
]



import struct
import numpy as np
import pytest
from timezonefinder.flatbuf.io.polygons import (
    get_polygon_collection,
    write_polygon_collection_flatbuffer,
    read_polygon_array_from_binary,
    flatten_polygon_coords,
    reshape_to_polygon_coords,
)
from timezonefinder.utils import close_resource


@pytest.mark.parametrize(
    "polygons",
    [
        [
            # Example polygons with separate x and y coordinates
            np.array([[0, 1, 2], [3, 4, 5]]),
            np.array([[6, 7, 8, 9], [10, 11, 12, 13]]),
        ],
    ],
)
def test_single_polygon_collection_round_trip(tmp_path, polygons):
    """Test that writing and reading a single polygon collection gives the same results."""
    # Define output path
    output_file = tmp_path / "polygons.fbs"

    # Write polygons to a single binary file
    write_polygon_collection_flatbuffer(output_file, polygons)

    assert output_file.exists(), "Output file should exist after writing."
    assert output_file.stat().st_size > 0, "Output file should be non-empty."

    with open(output_file, "rb") as file:
        # Read the binary file and verify the polygons
        buffer = file.read()
    poly_collection = get_polygon_collection(buffer)
    for idx, original_polygon in enumerate(polygons):
        read_polygon = read_polygon_array_from_binary(poly_collection, idx)
        np.testing.assert_array_equal(
            read_polygon, original_polygon, "Polygon mismatch."
        )

    with pytest.raises(struct.error):
        # Attempt to read a polygon with an out-of-bounds index
        read_polygon = read_polygon_array_from_binary(poly_collection, idx + 1)

    close_resource(buffer)
    close_resource(file)


@pytest.mark.parametrize(
    "polygon",
    [
        # Test with different shapes and values
        np.array([[0, 1, 2], [3, 4, 5]]),  # 2x3 array
        np.array([[6, 7, 8, 9], [10, 11, 12, 13]]),  # 2x4 array
        np.array([[100, 200], [300, 400]]),  # 2x2 array
        np.array([[1000], [2000]]),  # 2x1 array
        np.array(
            [[-10, -20, -30, -40, -50], [60, 70, 80, 90, 100]]
        ),  # 2x5 array with negative values
    ],
)
def test_coordinate_transformation_functions(polygon):
    """Test that flatten_polygon_coords and reshape_to_polygon_coords are inverses of each other."""
    # Test polygon -> flattened -> polygon round trip
    flattened = flatten_polygon_coords(polygon)
    reconstructed = reshape_to_polygon_coords(flattened)
    np.testing.assert_array_equal(
        reconstructed,
        polygon,
        "Reconstructed polygon does not match the original after flattening and reshaping.",
    )

    # Verify the flattened structure: [x0, y0, x1, y1, ...]
    expected_flat = np.array([])
    for i in range(polygon.shape[1]):
        expected_flat = np.append(expected_flat, [polygon[0, i], polygon[1, i]])

    np.testing.assert_array_equal(
        flattened,
        expected_flat,
        "Flattened array structure does not match expected [x0, y0, x1, y1, ...] pattern.",
    )

    # Verify dimensions
    assert flattened.shape[0] == polygon.shape[0] * polygon.shape[1], (
        "Flattened shape is incorrect"
    )
    assert reconstructed.shape == polygon.shape, (
        "Reconstructed shape does not match original"
    )


if __name__ == "__main__":
    pytest.main([__file__])



from pathlib import Path
from typing import List

from timezonefinder.configs import DEFAULT_DATA_DIR


def get_zone_names_path(output_path: Path = DEFAULT_DATA_DIR) -> Path:
    """Get the path to the timezone names text file."""
    return output_path / "timezone_names.txt"


def write_zone_names(
    zone_names: List[str], output_path: Path = DEFAULT_DATA_DIR
) -> None:
    """
    Write timezone names to a text file.

    Args:
        zone_names: List of timezone names.
        output_path: Directory where the output file will be written.
    """
    path = get_zone_names_path(output_path)
    with open(path, "w", encoding="utf-8") as f:
        f.write("\n".join(zone_names))
        f.write("\n")  # write a newline at the end of the file


def read_zone_names(path: Path) -> List[str]:
    """
    Read timezone names from a text file.

    Args:
        path: Path to the timezone names text file.
              If None, the default path will be used.

    Returns:
        List of timezone names.
    """
    file_path = get_zone_names_path(path)
    with open(file_path, encoding="utf-8") as f:
        return [line.strip() for line in f if line.strip()]



from timezonefinder.timezonefinder import (
    TimezoneFinder,
    TimezoneFinderL,
)

# Import module-level functions
from timezonefinder.global_functions import (
    timezone_at,
    timezone_at_land,
    unique_timezone_at,
    certain_timezone_at,
    get_geometry,
)

# https://docs.python.org/3/tutorial/modules.html#importing-from-a-package
# determines which objects will be imported with "import *"
__all__ = (
    "TimezoneFinder",
    "TimezoneFinderL",
    "timezone_at",
    "timezone_at_land",
    "unique_timezone_at",
    "certain_timezone_at",
    "get_geometry",
)



"""
Configuration file for pytest.
"""


def pytest_configure(config):
    """
    Register custom markers for different types of tests.
    """
    config.addinivalue_line("markers", "integration: mark test as integration test")
    config.addinivalue_line("markers", "unit: mark test as unit test")



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: shortcuts_uint16

import flatbuffers
from flatbuffers.compat import import_numpy

np = import_numpy()


class PolygonList:
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = PolygonList()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsPolygonList(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)

    # PolygonList
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # PolygonList
    def PolyIds(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint16Flags,
                a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 2),
            )
        return 0

    # PolygonList
    def PolyIdsAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint16Flags, o)
        return 0

    # PolygonList
    def PolyIdsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # PolygonList
    def PolyIdsIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        return o == 0


def PolygonListStart(builder):
    builder.StartObject(1)


def Start(builder):
    PolygonListStart(builder)


def PolygonListAddPolyIds(builder, polyIds):
    builder.PrependUOffsetTRelativeSlot(
        0, flatbuffers.number_types.UOffsetTFlags.py_type(polyIds), 0
    )


def AddPolyIds(builder, polyIds):
    PolygonListAddPolyIds(builder, polyIds)


def PolygonListStartPolyIdsVector(builder, numElems):
    return builder.StartVector(2, numElems, 2)


def StartPolyIdsVector(builder, numElems):
    return PolygonListStartPolyIdsVector(builder, numElems)


def PolygonListEnd(builder):
    return builder.EndObject()


def End(builder):
    return PolygonListEnd(builder)



"""optionally builds inside polygon algorithm C extension

Resources:
https://github.com/FirefoxMetzger/mini-extension
https://stackoverflow.com/questions/60073711/how-to-build-c-extensions-via-poetry
https://github.com/libmbd/libmbd/blob/master/build.py
"""

import pathlib
import re
from typing import Optional
import warnings

import cffi

EXTENSION_NAME = "inside_polygon_ext"
H_FILE_NAME = "inside_polygon_int.h"
C_FILE_NAME = "inside_polygon_int.c"
EXTENSION_PATH = pathlib.Path("timezonefinder") / "inside_poly_extension"
h_file_path = EXTENSION_PATH / H_FILE_NAME
c_file_path = EXTENSION_PATH / C_FILE_NAME

ffibuilder: Optional[cffi.FFI] = None
try:
    ffibuilder = cffi.FFI()
except Exception as exc:
    # Clang extension should be fully optional
    warnings.warn(
        f"C lang extension cannot be build, since cffi failed with this error: {exc}"
    )

if ffibuilder is not None:
    ffibuilder.set_source(
        "timezonefinder." + EXTENSION_NAME,
        source='#include "inside_polygon_int.h"',
        sources=[str(c_file_path)],
        include_dirs=[str(EXTENSION_PATH)],
    )
    with open(h_file_path) as h_file:
        # cffi does not like our preprocessor directives, so we remove them
        lns = h_file.read().splitlines()
        flt = filter(lambda ln: not re.match(r" *#", ln), lns)

    ffibuilder.cdef("\n".join(flt))


if __name__ == "__main__":
    if ffibuilder:
        ffibuilder.compile(verbose=True)



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: polygons

import flatbuffers
from flatbuffers.compat import import_numpy

np = import_numpy()


class Polygon:
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Polygon()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsPolygon(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)

    # Polygon
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Polygon
    def Coords(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Int32Flags,
                a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4),
            )
        return 0

    # Polygon
    def CoordsAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Int32Flags, o)
        return 0

    # Polygon
    def CoordsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Polygon
    def CoordsIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        return o == 0


def PolygonStart(builder):
    builder.StartObject(1)


def Start(builder):
    PolygonStart(builder)


def PolygonAddCoords(builder, coords):
    builder.PrependUOffsetTRelativeSlot(
        0, flatbuffers.number_types.UOffsetTFlags.py_type(coords), 0
    )


def AddCoords(builder, coords):
    PolygonAddCoords(builder, coords)


def PolygonStartCoordsVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)


def StartCoordsVector(builder, numElems):
    return PolygonStartCoordsVector(builder, numElems)


def PolygonEnd(builder):
    return builder.EndObject()


def End(builder):
    return PolygonEnd(builder)



from pathlib import Path
from typing import Iterable, Union

import numpy as np

from timezonefinder.configs import IntegerLike

from timezonefinder import utils
from timezonefinder.coord_accessors import AbstractCoordAccessor, create_coord_accessor
from timezonefinder.flatbuf.io.polygons import (
    get_coordinate_path,
)
from timezonefinder.np_binary_helpers import (
    get_xmax_path,
    get_xmin_path,
    get_ymax_path,
    get_ymin_path,
    read_per_polygon_vector,
)


class PolygonArray:
    xmin: np.ndarray
    xmax: np.ndarray
    ymin: np.ndarray
    ymax: np.ndarray
    coordinates: AbstractCoordAccessor

    def __init__(
        self,
        data_location: Union[str, Path],
        in_memory: bool = False,
    ):
        """
        Initialize the PolygonArray.
        :param data_location: The path to the binary data files to use.
        :param in_memory: Whether to completely read and keep the coordinate data in memory as numpy.
        """
        self.in_memory = in_memory
        self.data_location: Path = Path(data_location)

        xmin_path = get_xmin_path(self.data_location)
        xmax_path = get_xmax_path(self.data_location)
        ymin_path = get_ymin_path(self.data_location)
        ymax_path = get_ymax_path(self.data_location)

        # read all per polygon vectors directly into memory (no matter the memory mode)
        self.xmin = read_per_polygon_vector(xmin_path)
        self.xmax = read_per_polygon_vector(xmax_path)
        self.ymin = read_per_polygon_vector(ymin_path)
        self.ymax = read_per_polygon_vector(ymax_path)

        coordinate_file_path = get_coordinate_path(self.data_location)
        # Initialize the appropriate coordinate accessor based on memory mode
        self.coordinates = create_coord_accessor(coordinate_file_path, self.in_memory)

    def __del__(self):
        """Clean up resources when the object is destroyed."""
        del self.coordinates
        del self.xmin
        del self.xmax
        del self.ymin
        del self.ymax

    def __len__(self) -> int:
        """
        Get the number of polygons in the collection.
        :return: Number of polygons
        """
        return len(self.xmin)

    def outside_bbox(self, poly_id: IntegerLike, x: int, y: int) -> bool:
        """
        Check if a point is outside the bounding box of a polygon.

        :param poly_id: Polygon ID
        :param x: X-coordinate of the point
        :param y: Y-coordinate of the point
        :return: True if the point is outside the boundaries, False otherwise
        """
        if x > self.xmax[poly_id]:
            return True
        if x < self.xmin[poly_id]:
            return True
        if y > self.ymax[poly_id]:
            return True
        if y < self.ymin[poly_id]:
            return True
        return False

    def coords_of(self, idx: IntegerLike) -> np.ndarray:
        """
        Get the polygon coordinates for the given index.

        Args:
            idx: The polygon index

        Returns:
            A numpy array containing the polygon coordinates
        """
        return self.coordinates[idx]

    def pip(self, poly_id: IntegerLike, x: int, y: int) -> bool:
        """
        Point in polygon (PIP) test.

        :param poly_id: Polygon ID
        :param x: X-coordinate of the point
        :param y: Y-coordinate of the point
        :return: True if the point is inside the polygon, False otherwise
        """
        polygon = self.coords_of(poly_id)
        return utils.inside_polygon(x, y, polygon)

    def pip_with_bbox_check(self, poly_id: IntegerLike, x: int, y: int) -> bool:
        """
        Point in polygon (PIP) test with bounding box check.

        :param poly_id: Polygon ID
        :param x: X-coordinate of the point
        :param y: Y-coordinate of the point
        :return: True if the point is inside the polygon, False otherwise
        """
        if self.outside_bbox(poly_id, x, y):
            return False
        return self.pip(poly_id, x, y)

    def in_any_polygon(self, poly_ids: Iterable[int], x: int, y: int) -> bool:
        """
        Check if a point is inside any of the specified polygons.

        :param poly_ids: An iterable of polygon IDs
        :param x: X-coordinate of the point
        :param y: Y-coordinate of the point
        :return: True if the point is inside any polygon, False otherwise
        """
        for poly_id in poly_ids:
            if self.pip_with_bbox_check(poly_id, x, y):
                return True
        return False



"""utility functions"""

from pathlib import Path
import re
from typing import Any, Callable, Tuple

import numpy as np

from timezonefinder.configs import (
    DEFAULT_DATA_DIR,
    OCEAN_TIMEZONE_PREFIX,
)
from timezonefinder import utils_numba, utils_clang


# make numba functions available via utils
using_numba = utils_numba.using_numba
clang_extension_loaded = utils_clang.clang_extension_loaded
is_valid_lat = utils_numba.is_valid_lat
is_valid_lng = utils_numba.is_valid_lng
coord2int = utils_numba.coord2int
int2coord = utils_numba.int2coord
convert2coords = utils_numba.convert2coords
convert2coord_pairs = utils_numba.convert2coord_pairs
get_last_change_idx = utils_numba.get_last_change_idx


inside_polygon: Callable[[int, int, np.ndarray], bool]
# at import time fix which "point-in-polygon" implementation will be used
if clang_extension_loaded and not using_numba:
    # use the C implementation only if Numba is not present
    inside_polygon = utils_clang.pt_in_poly_clang
else:
    # use the (JIT compiled) python function if Numba is present or the C extension cannot be loaded
    inside_polygon = utils_numba.pt_in_poly_python


def validate_lat(lat: float) -> None:
    if not is_valid_lat(lat):
        raise ValueError(f"The given latitude {lat} is out of bounds")


def validate_lng(lng: float) -> None:
    if not is_valid_lng(lng):
        raise ValueError(f"The given longitude {lng} is out of bounds")


def validate_coordinates(lng: float, lat: float) -> Tuple[float, float]:
    lng, lat = float(lng), float(lat)
    validate_lng(lng)
    validate_lat(lat)
    return lng, lat


def close_resource(obj: Any) -> None:
    """Safely close a resource, ignoring specific expected errors."""
    if obj is None:
        return
    try:
        obj.close()
    except (AttributeError, OSError, ValueError):
        pass


def is_ocean_timezone(timezone_name: str) -> bool:
    if re.match(OCEAN_TIMEZONE_PREFIX, timezone_name) is None:
        return False
    return True


def get_boundaries_dir(data_dir: Path = DEFAULT_DATA_DIR) -> Path:
    """Return the path to the boundaries directory."""
    return data_dir / "boundaries"


def get_holes_dir(data_dir: Path = DEFAULT_DATA_DIR) -> Path:
    """Return the path to the holes directory."""
    return data_dir / "holes"


def get_hole_registry_path(data_dir: Path = DEFAULT_DATA_DIR) -> Path:
    """Return the path to the hole registry file."""
    return data_dir / "hole_registry.json"



import os
from pathlib import Path
from typing import Any, Dict, List, Tuple, Union

import numpy as np

# SHORTCUT SETTINGS
# h3 library
SHORTCUT_H3_RES: int = 3

OCEAN_TIMEZONE_PREFIX = r"Etc/GMT"

# PATHS
PACKAGE_DIR = Path(__file__).parent
DEFAULT_DATA_DIR = PACKAGE_DIR / "data"


# i = signed 4byte integer
NR_BYTES_I = 4
# IMPORTANT: all values between -180 and 180 degree must fit into the domain of i4!
# is the same as testing if 360 fits into the domain of I4 (unsigned!)
MAX_ALLOWED_COORD_VAL = 2 ** (8 * NR_BYTES_I - 1)

# from math import floor,log10
# DECIMAL_PLACES_SHIFT = floor(log10(MAX_ALLOWED_COORD_VAL/180.0)) # == 7
DECIMAL_PLACES_SHIFT = 7
INT2COORD_FACTOR = 10 ** (-DECIMAL_PLACES_SHIFT)
COORD2INT_FACTOR = 10**DECIMAL_PLACES_SHIFT
MAX_LNG_VAL = 180.0
MAX_LAT_VAL = 90.0
MAX_LNG_VAL_INT = int(MAX_LNG_VAL * COORD2INT_FACTOR)
MAX_LAT_VAL_INT = int(MAX_LAT_VAL * COORD2INT_FACTOR)
MAX_INT_VAL = MAX_LNG_VAL_INT
assert MAX_INT_VAL < MAX_ALLOWED_COORD_VAL

# TYPES
# used in Numba JIT compiled function signatures in utils_numba.py
# NOTE: Changes in the global settings might not immediately affect
#  the functions due to caching!

# Type alias for flexibility with integer types (pure int or numpy integer scalars)
IntegerLike = Union[int, np.integer]

# hexagon id to list of polygon ids
ShortcutMapping = Dict[int, np.ndarray]
CoordPairs = List[Tuple[float, float]]
CoordLists = List[List[float]]
IntLists = List[List[int]]


# zone id storage settings ---------------------------------------------------

_ZONE_ID_DTYPE_ALIASES: Dict[str, "np.dtype[Any]"] = {
    "uint8": np.dtype("<u1"),
    "uint16": np.dtype("<u2"),
}


def _normalise_zone_id_dtype_key(key: str) -> str:
    """Normalise user provided dtype keys to canonical form."""
    return key.lower().strip()


def get_zone_id_dtype(name: str) -> "np.dtype[Any]":
    """Return the configured numpy dtype for storing zone IDs."""

    try:
        return _ZONE_ID_DTYPE_ALIASES[_normalise_zone_id_dtype_key(name)]
    except KeyError as exc:  # pragma: no cover - defensive, validated on import
        valid = ", ".join(sorted(_ZONE_ID_DTYPE_ALIASES))
        raise ValueError(
            f"Unsupported zone id dtype '{name}'. Choose one of: {valid}"
        ) from exc


def zone_id_dtype_to_string(dtype: np.dtype) -> str:
    """Return the little-endian numpy dtype string for serialisation."""

    return dtype.newbyteorder("<").str


def available_zone_id_dtype_names() -> Tuple[str, ...]:
    """Return the supported zone id dtype names."""

    return tuple(sorted(_ZONE_ID_DTYPE_ALIASES))


DEFAULT_ZONE_ID_DTYPE_NAME = os.getenv("TIMEZONEFINDER_ZONE_ID_DTYPE", "uint8")
DEFAULT_ZONE_ID_DTYPE = get_zone_id_dtype(DEFAULT_ZONE_ID_DTYPE_NAME)



"""
Utility functions for handling .npy numpy binary files related to timezone data.
"""

from pathlib import Path

import numpy as np


def get_zone_ids_path(path: Path) -> Path:
    """Return the path to the zone_ids.npy file in the given directory."""
    return path / "zone_ids.npy"


def get_zone_positions_path(path: Path) -> Path:
    """Return the path to the zone_positions.npy file in the given directory."""
    return path / "zone_positions.npy"


def get_xmax_path(path: Path) -> Path:
    """Return the path to the xmax.npy file in the given directory."""
    return path / "xmax.npy"


def get_xmin_path(path: Path) -> Path:
    """Return the path to the xmin.npy file in the given directory."""
    return path / "xmin.npy"


def get_ymax_path(path: Path) -> Path:
    """Return the path to the ymax.npy file in the given directory."""
    return path / "ymax.npy"


def get_ymin_path(path: Path) -> Path:
    """Return the path to the ymin.npy file in the given directory."""
    return path / "ymin.npy"


def store_per_polygon_vector(file_path: Path, vector: np.ndarray) -> None:
    """Store a vector as a .npy file in the specified file path."""
    print(f"Storing vector to {file_path}")
    np.save(file_path, vector)


def read_per_polygon_vector(file_path: Path) -> np.ndarray:
    """Read a vector from a .npy file in the specified file path."""
    vector = np.load(file_path)
    return vector



import flatbuffers
import mmap
import numpy as np
from pathlib import Path
from typing import List, Union

from timezonefinder.configs import DEFAULT_DATA_DIR
from timezonefinder.flatbuf.generated.polygons.Polygon import (
    PolygonStart,
    PolygonEnd,
    PolygonAddCoords,
    PolygonStartCoordsVector,
)
from timezonefinder.flatbuf.generated.polygons.PolygonCollection import (
    PolygonCollection,
    PolygonCollectionStart,
    PolygonCollectionEnd,
    PolygonCollectionAddPolygons,
    PolygonCollectionStartPolygonsVector,
)


def flatten_polygon_coords(polygon: np.ndarray) -> np.ndarray:
    """Convert polygon coordinates from shape (2, N) to a flattened [x0, y0, x1, y1, ...] array.

    Args:
        polygon: Array of polygon coordinates with shape (2, N)
                where the first row contains x coordinates and the second row contains y coordinates

    Returns:
        Flattened 1D array of coordinates in the format [x0, y0, x1, y1, ...]
    """
    return polygon.ravel(order="F")


def reshape_to_polygon_coords(coords: np.ndarray) -> np.ndarray:
    """Reshape flattened coordinates to the format (2, N).

    Args:
        coords: Flattened 1D array of coordinates in the format [x0, y0, x1, y1, ...]

    Returns:
        Array of polygon coordinates with shape (2, N)
        where the first row contains x coordinates and the second row contains y coordinates
    """
    return coords.reshape(2, -1, order="F")


def get_coordinate_path(data_dir: Path = DEFAULT_DATA_DIR) -> Path:
    """Return the path to the boundaries flatbuffer file."""
    return data_dir / "coordinates.fbs"


def write_polygon_collection_flatbuffer(
    file_path: Path, polygons: List[np.ndarray]
) -> None:
    """Write a collection of polygons to a flatbuffer file using a single coordinate vector.

    Args:
        file_path: Path to save the flatbuffer file
        polygons: List of polygon coordinates as numpy arrays with shape (2, N)
                  where the first row contains x coordinates and the second row contains y coordinates

    Returns:
        None
    """
    print(f"writing {len(polygons)} polygons to binary file {file_path}")
    builder = flatbuffers.Builder(0)
    polygon_offsets = []

    # Create each polygon and store its offset
    for polygon in polygons:
        # Flatten coordinates to [x0, y0, x1, y1, ...] format
        coords = flatten_polygon_coords(polygon)

        # Create coords vector
        PolygonStartCoordsVector(builder, len(coords))
        for coord in reversed(coords):
            builder.PrependInt32(int(coord))  # Use signed 32-bit integer
        coords_offset = builder.EndVector()

        # Create polygon
        PolygonStart(builder)
        PolygonAddCoords(builder, coords_offset)  # Use Coords for combined vector
        polygon_offsets.append(PolygonEnd(builder))

    # Create polygon vector
    PolygonCollectionStartPolygonsVector(builder, len(polygon_offsets))
    for offset in reversed(polygon_offsets):
        builder.PrependUOffsetTRelative(offset)
    polygons_offset = builder.EndVector()

    # Create root table
    PolygonCollectionStart(builder)
    PolygonCollectionAddPolygons(builder, polygons_offset)
    collection_offset = PolygonCollectionEnd(builder)

    # Finish buffer
    builder.Finish(collection_offset)

    # Write to file
    with open(file_path, "wb") as f:
        buf = builder.Output()
        f.write(buf)


def get_polygon_collection(buf: Union[bytes, mmap.mmap]) -> PolygonCollection:
    """Load a PolygonCollection from a file path.

    Args:
        buf: A binary stream or memory-mapped file containing the flatbuffer data.

    Returns: PolygonCollection
    """
    return PolygonCollection.GetRootAs(buf, 0)


def read_polygon_array_from_binary(
    poly_collection: PolygonCollection, idx: int
) -> np.ndarray:
    """Read a polygon's coordinates from a FlatBuffers collection."""
    # value checks not required as this is a private function
    # processed polygon indices are expected to be in range
    # nr_polygons = collection.PolygonsLength()
    # if idx >= nr_polygons:
    #     raise IndexError(
    #         f"Index {idx} out of bounds for collection with {nr_polygons} polygons."
    #     )
    poly = poly_collection.Polygons(idx)
    coords = poly.CoordsAsNumpy()  # flat 1D array of coordinates
    # Reshape to (2, N) format
    return reshape_to_polygon_coords(coords)



import subprocess
import sys
from pathlib import Path
import pytest


def get_example_scripts():
    examples_dir = Path(__file__).parent.parent / "examples"
    scripts = []
    for file_path in examples_dir.iterdir():
        if (
            file_path.is_file()
            and file_path.suffix == ".py"
            and not file_path.name.startswith("__")
        ):
            scripts.append(file_path.stem)
    return scripts


@pytest.mark.parametrize(
    "script_name",
    get_example_scripts(),
)
def test_example_script_runs(script_name):
    module_name = f"examples.{script_name}"
    print(f"Running {module_name} script:")
    result = subprocess.run(
        [sys.executable, "-m", module_name],
        capture_output=True,
        text=True,
    )
    print(result.stdout)
    print(result.stderr)
    assert result.returncode == 0, f"Script failed: {result.stderr}"



from importlib.util import find_spec
from typing import Optional

import numpy as np
import pytest

from scripts.configs import THRES_DTYPE_H
from tests.auxiliaries import (
    check_geometry,
    check_pairwise_geometry,
    ocean2land,
    validate_polygon_coordinates,
)
from tests.global_functions_test import single_location_test
from tests.locations import BASIC_TEST_LOCATIONS, EDGE_TEST_CASES, TEST_LOCATIONS
from timezonefinder.configs import (
    INT2COORD_FACTOR,
)
from timezonefinder.polygon_array import PolygonArray
from timezonefinder.timezonefinder import (
    AbstractTimezoneFinder,
    TimezoneFinder,
    TimezoneFinderL,
)
from timezonefinder.utils import get_boundaries_dir, is_ocean_timezone

DEBUG = False
# more extensive testing (e.g. get geometry for every single zone), switch off for CI/CD
# DEBUG = True

PACKAGE_NAME = "timezonefinder"

boundaries_dir = get_boundaries_dir()
boundaries = PolygonArray(data_location=boundaries_dir, in_memory=False)
NR_TZ_POLYGONS = len(boundaries)

NR_STARTUPS_PER_CLASS = 1

class_under_test = TimezoneFinder
tf: AbstractTimezoneFinder = class_under_test()
in_memory_mode = False

RESULT_TEMPLATE = "{0:25s} | {1:20s} | {2:20s} | {3:2s}"


# tests for both classes: TimezoneFinderL and TimezoneFinder
class TestBaseTimezoneFinderClass:
    class_under_test = TimezoneFinderL
    # NOTE: setting memory mode does not make a difference for TimezoneFinderL (relevant only for polygon data)
    in_memory_mode = False
    bin_file_dir = None
    on_land_pt_fct_name = "timezone_at"
    test_locations = BASIC_TEST_LOCATIONS

    def test_using_numba(self):
        spec = find_spec("numba")
        numba_installed = spec is not None
        assert self.test_instance.using_numba() == numba_installed

    def test_using_clang_pip(self):
        res = self.test_instance.using_clang_pip()
        assert isinstance(res, bool)

    def print_tf_class_props(self):
        print("test properties:")
        print(f"testing class {self.class_under_test}")
        print(
            f"using_numba()=={self.class_under_test.using_numba()} (JIT compiled functions {'NOT ' if not self.class_under_test.using_numba() else ''}in use)"
        )
        print(f"in_memory={self.in_memory_mode}")
        print(f"file location={self.bin_file_dir}\n")

    @classmethod
    def setup_class(cls):
        # preparations which have to be made only once
        cls.print_tf_class_props(cls)
        cls.test_instance = cls.class_under_test(
            bin_file_location=cls.bin_file_dir, in_memory=cls.in_memory_mode
        )

    def check_timezone_at_results(self, lng, lat, expected: Optional[str] = ""):
        # at the edges of the coordinate system the algorithms should still be well defined!

        print(
            [
                self.test_instance.zone_name_from_boundary_id(b_id)
                for b_id in self.test_instance._iter_boundaries_in_shortcut(
                    lng=lng, lat=lat
                )
            ]
        )

        result = self.test_instance.timezone_at(lng=lng, lat=lat)
        if isinstance(expected, str) and len(expected) == 0:
            # zone_name="" is interpreted as "don't care"
            return
        assert result == expected

    def test_edge_shortcut_validity(self):
        for lng, lat, _expected in EDGE_TEST_CASES:
            self.check_timezone_at_results(lng, lat)

        with pytest.raises(ValueError):
            self.check_timezone_at_results(lng=180.0 + INT2COORD_FACTOR, lat=90.0)
            self.check_timezone_at_results(
                lng=-180.0 - INT2COORD_FACTOR, lat=90.0 + INT2COORD_FACTOR
            )
            self.check_timezone_at_results(lng=-180.0, lat=90.0 + INT2COORD_FACTOR)
            self.check_timezone_at_results(lng=180.0 + INT2COORD_FACTOR, lat=-90.0)
            self.check_timezone_at_results(lng=180.0, lat=-90.0 - INT2COORD_FACTOR)
            self.check_timezone_at_results(lng=-180.0 - INT2COORD_FACTOR, lat=-90.0)
            self.check_timezone_at_results(
                lng=-180.0 - INT2COORD_FACTOR, lat=-90.01 - INT2COORD_FACTOR
            )

    def test_kwargs_only(self):
        # calling timezonefinder fcts without keyword arguments should raise an error
        with pytest.raises(TypeError):
            self.test_instance.timezone_at(23.0, 42.0)
            self.test_instance.timezone_at(23.0, lng=42.0)
            self.test_instance.timezone_at(23.0, lat=42.0)

            self.test_instance.timezone_at_land(23.0, 42.0)
            self.test_instance.timezone_at_land(23.0, lng=42.0)
            self.test_instance.timezone_at_land(23.0, lat=42.0)

    @staticmethod
    def run_location_tests(test_fct, lat, lng, loc, expected_orig):
        single_location_test(test_fct, lat, lng, loc, expected_orig)

    @classmethod
    def pytest_generate_tests(cls, metafunc):
        # Dynamically generate test parameters based on the test method
        if metafunc.function.__name__ == "test_timezone_at":
            metafunc.parametrize("lat, lng, loc, expected", cls.test_locations)
        elif metafunc.function.__name__ == "test_timezone_at_land":
            metafunc.parametrize(
                "lat, lng, loc, expected", list(ocean2land(cls.test_locations))
            )
        elif metafunc.function.__name__ == "test_unambiguous_timezone_at":
            metafunc.parametrize("lat, lng, loc, expected", BASIC_TEST_LOCATIONS)

    def test_timezone_at(self, lat, lng, loc, expected):
        self.run_location_tests(self.test_instance.timezone_at, lat, lng, loc, expected)

    def test_timezone_at_land(self, lat, lng, loc, expected):
        self.run_location_tests(
            self.test_instance.timezone_at_land, lat, lng, loc, expected
        )

    def test_unambiguous_timezone_at(self, lat, lng, loc, expected):
        self.run_location_tests(
            self.test_instance.unique_timezone_at, lat, lng, loc, expected
        )

    def test_timezone_names(self):
        timezone_names_stored = self.test_instance.timezone_names
        assert isinstance(timezone_names_stored, list)
        assert len(timezone_names_stored) > 0, "no timezone names found"
        # test if all timezone names are strings
        assert all(isinstance(name, str) for name in timezone_names_stored), (
            "not all timezone names are strings"
        )
        # test if all timezone names are unique
        assert len(set(timezone_names_stored)) == len(timezone_names_stored), (
            "not all timezone names are unique"
        )
        # test if all timezone names are valid
        for name in timezone_names_stored:
            assert len(name) > 0, f"empty timezone name found: {name}"
            assert "/" in name or is_ocean_timezone(name), (
                f"invalid timezone name: {name}. It should contain a '/' or be an ocean timezone."
            )

            # TODO further checks for valid timezone names


# tests for Timezonefinder class
class TestTimezonefinderClass(TestBaseTimezoneFinderClass):
    class_under_test = TimezoneFinder
    on_land_pt_fct_name = "timezone_at_land"
    test_locations = TEST_LOCATIONS

    def test_kwargs_only(self):
        super().test_kwargs_only()

        with pytest.raises(TypeError):
            self.test_instance.certain_timezone_at(23.0, 42.0)
            self.test_instance.certain_timezone_at(23.0, lng=42.0)
            self.test_instance.certain_timezone_at(23.0, lat=42.0)

    def test_nr_of_polygons(self):
        res = self.test_instance.nr_of_polygons
        assert isinstance(res, int)
        assert res > 0
        assert res < THRES_DTYPE_H

    # test if all polygon coordinates can be retrieved
    # NOTE: too many polygons, so this test is not parametrized
    def test_coords_of(self):
        nr_of_polygons = self.test_instance.nr_of_polygons
        for poly_id in range(nr_of_polygons):
            print(f"Testing polygon ID: {poly_id}")
            coords = self.test_instance.coords_of(poly_id)
            validate_polygon_coordinates(coords)

    def test_holes_of_poly(self):
        print("test retrieving all holes for each polygon using _holes_of_poly:")
        nr_of_polygons = self.test_instance.nr_of_polygons
        for poly_id in range(nr_of_polygons):
            print(f"polygon ID: {poly_id + 1}/{nr_of_polygons}", end="\r", flush=True)
            for i, hole_coords in enumerate(self.test_instance._holes_of_poly(poly_id)):
                print(
                    f"polygon ID: {poly_id + 1}/{nr_of_polygons}, hole {i + 1}",
                    end="\r",
                    flush=True,
                )
                validate_polygon_coordinates(hole_coords)
        print()  # move to next line after progress output

    def test_edge_shortcut_result(self):
        for lng, lat, expected in EDGE_TEST_CASES:
            # NOTE: for TimezoneFinder (using polygon data) the results must match!
            self.check_timezone_at_results(lng, lat, expected)

    def test_certain_timezone_at(self, lat, lng, loc, expected):
        self.run_location_tests(
            self.test_instance.certain_timezone_at, lat, lng, loc, expected
        )

    @classmethod
    def pytest_generate_tests(cls, metafunc):
        # call the super class method
        super().pytest_generate_tests(metafunc)
        if metafunc.function.__name__ == "test_certain_timezone_at" and hasattr(
            cls, "test_certain_timezone_at"
        ):
            metafunc.parametrize("lat, lng, loc, expected", cls.test_locations)

    def test_overflow(self):
        longitude = -123.2
        latitude = 48.4
        # make numpy overflow runtime warning raise an error

        np.seterr(all="warn")
        import warnings

        warnings.filterwarnings("error")
        # must not raise a warning
        self.test_instance.certain_timezone_at(
            lat=float(latitude), lng=float(longitude)
        )

    def test_get_geometry(self):
        print("testing get_geometry():")
        timezone_names_stored = self.test_instance.timezone_names
        nr_timezones = len(timezone_names_stored)
        for zone_id, zone_name in enumerate(timezone_names_stored):
            if not DEBUG and zone_id > 5:
                break

            print(zone_id, zone_name)
            geometry_from_name = self.test_instance.get_geometry(
                tz_name=zone_name, tz_id=None, use_id=False, coords_as_pairs=False
            )
            check_geometry(geometry_from_name)

            # conduct extensive testing only with active debugging
            geometry_from_id = self.test_instance.get_geometry(
                tz_name=zone_name,
                tz_id=zone_id,
                use_id=False,
                coords_as_pairs=False,
            )
            # not necessary:
            # assert nested_list_equal(geometry_from_id, geometry_from_name), \
            assert len(geometry_from_name) == len(geometry_from_id), (
                "the results for querying the geometry for a zone with zone name or zone id are NOT equal."
            )
            check_geometry(geometry_from_id)

            geometry_from_name = self.test_instance.get_geometry(
                tz_name=zone_name, tz_id=None, use_id=False, coords_as_pairs=True
            )
            geometry_from_id = self.test_instance.get_geometry(
                tz_name=zone_name, tz_id=zone_id, use_id=False, coords_as_pairs=True
            )
            assert len(geometry_from_name) == len(geometry_from_id), (
                "the results for querying the geometry for a zone with zone name or zone id are NOT equal."
            )

            check_pairwise_geometry(geometry_from_id)
            check_pairwise_geometry(geometry_from_name)

        with pytest.raises(ValueError):
            self.test_instance.get_geometry(
                tz_name="", tz_id=None, use_id=False, coords_as_pairs=False
            )
            self.test_instance.get_geometry(
                tz_name="", tz_id=0, use_id=False, coords_as_pairs=False
            )
            self.test_instance.get_geometry(
                tz_name="wrong_tz_name", tz_id=None, use_id=False, coords_as_pairs=False
            )
            self.test_instance.get_geometry(
                tz_name="wrong_tz_name", tz_id=0, use_id=False, coords_as_pairs=False
            )
            # id does not exist
            self.test_instance.get_geometry(
                tz_name=None, tz_id=nr_timezones, use_id=True, coords_as_pairs=False
            )
            self.test_instance.get_geometry(
                tz_name="", tz_id=-1, use_id=True, coords_as_pairs=False
            )


class TestTimezonefinderClassTestMEM(TestTimezonefinderClass):
    in_memory_mode = True


# TEST equality for all results. in_memory_mode = True/False must not change the results
# TEST equality for all results. in_memory_mode = True/False must not change the results



from typing import Final, Optional

import cffi

import numpy as np

ffi: Optional[cffi.FFI] = None
try:
    # Note: IDE might complain as this import comes from a cffi C extension
    from timezonefinder import inside_polygon_ext  # type: ignore

    clang_extension_loaded = True
    ffi = cffi.FFI()

except ImportError:
    clang_extension_loaded = False
    inside_polygon_ext = None

INT_LIST_REP: Final[str] = "int []"


def pt_in_poly_clang(x: int, y: int, coords: np.ndarray) -> bool:
    """wrapper of the point in polygon test algorithm C extension

    ATTENTION: the input numpy arrays must have a C_CONTIGUOUS memory layout
    https://numpy.org/doc/stable/reference/generated/numpy.ascontiguousarray.html?highlight=ascontiguousarray#numpy.ascontiguousarray
    """
    if ffi is None:
        raise ValueError(
            "Trying to use the clang implementation of the point in polygon algorithm "
            "while the C extension in not loaded."
        )
    x_coords = coords[0]
    y_coords = coords[1]
    nr_coords = len(x_coords)

    y_coords = np.ascontiguousarray(y_coords)
    x_coords = np.ascontiguousarray(x_coords)
    x_coords_ffi = ffi.from_buffer(INT_LIST_REP, x_coords)
    y_coords_ffi = ffi.from_buffer(INT_LIST_REP, y_coords)
    contained = inside_polygon_ext.lib.inside_polygon_int(
        x, y, nr_coords, x_coords_ffi, y_coords_ffi
    )
    return contained



import subprocess

import pytest

from timezonefinder.configs import DEFAULT_DATA_DIR
from timezonefinder.zone_names import read_zone_names

timezone_names = read_zone_names(DEFAULT_DATA_DIR)


@pytest.mark.parametrize(
    "cmd",
    [
        "timezonefinder 40.5 11.7",
        "timezonefinder -f 0 40.5 11.7",
        "timezonefinder -f 1 40.5 11.7",
        "timezonefinder -f 3 40.5 11.7",
        "timezonefinder -f 4 40.5 11.7",
        "timezonefinder -f 5 40.5 11.7",
    ],
)
def test_main(cmd: str):
    res = subprocess.getoutput(cmd).rstrip("\n\x1b[0m")
    assert not res.endswith("command not found"), "package not installed"
    assert res == "None" or res in timezone_names



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: shortcuts_uint16

import flatbuffers
from flatbuffers.compat import import_numpy

np = import_numpy()


class HybridShortcutEntry:
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = HybridShortcutEntry()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsHybridShortcutEntry(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)

    # HybridShortcutEntry
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # HybridShortcutEntry
    def HexId(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint64Flags, o + self._tab.Pos
            )
        return 0

    # HybridShortcutEntry
    def ValueType(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 0

    # HybridShortcutEntry
    def Value(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            from flatbuffers.table import Table

            obj = Table(bytearray(), 0)
            self._tab.Union(obj, o)
            return obj
        return None


def HybridShortcutEntryStart(builder):
    builder.StartObject(3)


def Start(builder):
    HybridShortcutEntryStart(builder)


def HybridShortcutEntryAddHexId(builder, hexId):
    builder.PrependUint64Slot(0, hexId, 0)


def AddHexId(builder, hexId):
    HybridShortcutEntryAddHexId(builder, hexId)


def HybridShortcutEntryAddValueType(builder, valueType):
    builder.PrependUint8Slot(1, valueType, 0)


def AddValueType(builder, valueType):
    HybridShortcutEntryAddValueType(builder, valueType)


def HybridShortcutEntryAddValue(builder, value):
    builder.PrependUOffsetTRelativeSlot(
        2, flatbuffers.number_types.UOffsetTFlags.py_type(value), 0
    )


def AddValue(builder, value):
    HybridShortcutEntryAddValue(builder, value)


def HybridShortcutEntryEnd(builder):
    return builder.EndObject()


def End(builder):
    return HybridShortcutEntryEnd(builder)



"""
Integration tests for the timezonefinder package installation and functionality.

This module contains tests that verify the correct installation and basic functionality
of the package through different distribution methods:
- Installation from a wheel package
- Installation from a source distribution (sdist)
- Installation directly from the source code

Each test creates a fresh virtual environment, installs the package using the specified
method, and then performs a basic test to ensure that the package can be imported and
that its core functionality works properly.
"""

import sys
import tempfile
from pathlib import Path
from typing import Tuple
import pytest

from scripts.configs import PROJECT_ROOT

from tests.auxiliaries import build_sdist, build_wheel, run_command

# Mark all tests in this module as integration tests
pytestmark = pytest.mark.integration

# Define additional path constants
BIN_DIR = "Scripts" if sys.platform == "win32" else "bin"


def run_timezonefinder_test(python_bin: str) -> None:
    """Test importing and instantiating TimezoneFinder."""
    code = "from timezonefinder import TimezoneFinder; tf = TimezoneFinder(); print(tf.timezone_at(lat=40.5, lng=11.7))"
    run_command([python_bin, "-c", code])


def setup_venv(tempdir: str) -> Tuple[str, str]:
    """Set up a virtual environment and return paths to python and pip binaries."""
    venv_dir = Path(tempdir) / "venv"
    run_command([sys.executable, "-m", "venv", str(venv_dir)])

    python_bin = str(venv_dir / BIN_DIR / "python")
    pip_bin = str(venv_dir / BIN_DIR / "pip")

    # Upgrade pip
    run_command([python_bin, "-m", "pip", "install", "--upgrade", "pip"])

    return python_bin, pip_bin


def check_install_package(package_path: Path) -> None:
    """
    Generic test function for installing and testing timezonefinder.

    Args:
        package_path: path to project for direct installation
    """
    with tempfile.TemporaryDirectory() as tempdir:
        temp_path = Path(tempdir)
        # Setup virtual environment
        python_bin, pip_bin = setup_venv(str(temp_path))

        # Install the package
        run_command([pip_bin, "install", str(package_path)])

        run_timezonefinder_test(python_bin)


def test_install_from_wheel():
    package_path = build_wheel()
    check_install_package(
        package_path,
    )


def test_install_from_sdist():
    package_path = build_sdist()
    check_install_package(
        package_path,
    )


def test_install_from_source():
    check_install_package(PROJECT_ROOT)



"""
Tests for the global functions in timezonefinder
"""

import pytest

from tests.auxiliaries import single_location_test
from tests.locations import BASIC_TEST_LOCATIONS, TEST_LOCATIONS_AT_LAND, TEST_LOCATIONS
from timezonefinder import (
    timezone_at,
    timezone_at_land,
    unique_timezone_at,
    certain_timezone_at,
    get_geometry,
    TimezoneFinder,
)
from timezonefinder.configs import DEFAULT_DATA_DIR
from timezonefinder.zone_names import read_zone_names

# Load all timezone names for parameterization
all_timezone_names = read_zone_names(DEFAULT_DATA_DIR)

FUNC2TEST_CASES = {
    timezone_at: TEST_LOCATIONS,
    certain_timezone_at: TEST_LOCATIONS,
    unique_timezone_at: BASIC_TEST_LOCATIONS,
    timezone_at_land: TEST_LOCATIONS_AT_LAND,
}


# Create parameterized test data from FUNC2TEST_CASES mapping
def create_test_params():
    """Create test parameters from FUNC2TEST_CASES mapping"""
    params = []
    ids = []
    for func, test_cases in FUNC2TEST_CASES.items():
        for lat, lng, description, expected in test_cases:
            params.append((func, lat, lng, description, expected))
            ids.append(f"{func.__name__}-{description}")
    return params, ids


# Generate test parameters and IDs once
TEST_PARAMS, TEST_IDS = create_test_params()


@pytest.mark.unit
class TestGlobalFunctions:
    """Test the global functions that use the global TimezoneFinder instance"""

    @pytest.mark.parametrize(
        "func, lat, lng, description, expected", TEST_PARAMS, ids=TEST_IDS
    )
    def test_global_functions_parameterized(
        self, func, lat, lng, description, expected
    ):
        """Test global functions with their respective test cases defined in FUNC2TEST_CASES"""
        single_location_test(func, lat, lng, description, expected)

    @pytest.mark.parametrize("tz_name", all_timezone_names)
    def test_get_geometry(self, tz_name):
        """Test the global get_geometry function for all timezones"""
        tf = TimezoneFinder()
        expected = tf.get_geometry(tz_name=tz_name)
        result = get_geometry(tz_name=tz_name)
        assert isinstance(result, type(expected)), (
            f"Type mismatch for {tz_name}: {type(result)} != {type(expected)}"
        )
        assert len(result) == len(expected), (
            f"Length mismatch for {tz_name}: {len(result)} != {len(expected)}"
        )



from timezonefinder.utils import is_ocean_timezone

# mapping from original timezone to reduced timezone of the timezones-now dataset
REDUCED_TIMEZONE_MAPPING: set[str, str] = {
    "Europe/Berlin": "Europe/Paris",
    "Europe/Kyiv": "Europe/Athens",
    "Asia/Yakutsk": "Asia/Tokyo",
    "Asia/Macau": "Asia/Manila",
    "Asia/Yekaterinburg": "Asia/Karachi",
    "Europe/Helsinki": "Europe/Athens",
    "Asia/Brunei": "Asia/Manila",
    "Asia/Nicosia": "Europe/Athens",
    "Asia/Seoul": "Asia/Tokyo",
    "Asia/Vladivostok": "Australia/Brisbane",
    "America/Inuvik": "America/Denver",
    "Asia/Anadyr": "Pacific/Fiji",
    "Asia/Yerevan": "Asia/Dubai",
    "Asia/Dushanbe": "Asia/Karachi",
    "Europe/Busingen": "Europe/Paris",
    "Europe/Zurich": "Europe/Paris",
    "Africa/Maseru": "Africa/Johannesburg",
    "Asia/Tashkent": "Asia/Karachi",
}


# for TimezoneFinderL:
BASIC_TEST_LOCATIONS = [
    # lat, lng, description, expected
    (35.295953, -89.662186, "Arlington, TN", "America/Chicago"),
    (35.1322601, -90.0902499, "Memphis, TN", "America/Chicago"),
    (61.17, -150.02, "Anchorage, AK", "America/Anchorage"),
    (40.2, -119.3, "California/Nevada border", "America/Los_Angeles"),
    (42.652647, -73.756371, "Albany, NY", "America/New_York"),
    (55.743749, 37.6207923, "Moscow", "Europe/Moscow"),
    (34.104255, -118.4055591, "Los Angeles", "America/Los_Angeles"),
    (55.743749, 37.6207923, "Moscow", "Europe/Moscow"),
    (39.194991, -106.8294024, "Aspen, Colorado", "America/Denver"),
    (50.438114, 30.5179595, "Kiev", "Europe/Kyiv"),
    (12.936873, 77.6909136, "Jogupalya", "Asia/Kolkata"),
    (38.889144, -77.0398235, "Washington DC", "America/New_York"),
    (19, -135, "pacific ocean", "Etc/GMT+9"),
    (30, -33, "atlantic ocean", "Etc/GMT+2"),
    (-24, 79, "indian ocean", "Etc/GMT-5"),
]


# for TimezoneFinder:
# certain algorithm should give the same results for all normal test cases
TEST_LOCATIONS = BASIC_TEST_LOCATIONS + [
    (59.932490, 30.3164291, "St Petersburg", "Europe/Moscow"),
    (50.300624, 127.559166, "Blagoveshchensk", "Asia/Yakutsk"),
    (42.439370, -71.0700416, "Boston", "America/New_York"),
    (41.84937, -87.6611995, "Chicago", "America/Chicago"),
    (28.626873, -81.7584514, "Orlando", "America/New_York"),
    (47.610615, -122.3324847, "Seattle", "America/Los_Angeles"),
    (51.499990, -0.1353549, "London", "Europe/London"),
    (51.256241, -0.8186531, "Church Crookham", "Europe/London"),
    (51.292215, -0.8002638, "Fleet", "Europe/London"),
    (48.868743, 2.3237586, "Paris", "Europe/Paris"),
    (22.158114, 113.5504603, "Macau", "Asia/Macau"),
    (56.833123, 60.6097054, "Russia", "Asia/Yekaterinburg"),
    (60.887496, 26.6375756, "Salo", "Europe/Helsinki"),
    (52.799992, -1.8524408, "Staffordshire", "Europe/London"),
    (5.016666, 115.0666667, "Muara", "Asia/Brunei"),
    (-41.466666, -72.95, "Puerto Montt seaport", "America/Santiago"),
    (34.566666, 33.0333333, "Akrotiri seaport", "Asia/Nicosia"),
    (37.466666, 126.6166667, "Inchon seaport", "Asia/Seoul"),
    (42.8, 132.8833333, "Nakhodka seaport", "Asia/Vladivostok"),
    (50.26, -5.051, "Truro", "Europe/London"),
    (37.790792, -122.389980, "San Francisco", "America/Los_Angeles"),
    (37.81, -122.35, "San Francisco Bay", "America/Los_Angeles"),
    (68.3597987, -133.745786, "America", "America/Inuvik"),
    # lng 180 == -180
    # 180.0: right on the timezone boundary polygon edge, the return value is uncertain (None in this case)
    # being tested in test_helpers.py
    (65.2, 179.9999, "lng 180", "Asia/Anadyr"),
    (65.2, -179.9999, "lng -180", "Asia/Anadyr"),
    # test cases for hole handling:
    (41.0702284, 45.0036352, "Aserbaid. Enklave", "Asia/Yerevan"),
    (39.8417402, 70.6020068, "Tajikistani Enklave", "Asia/Dushanbe"),
    (47.7024174, 8.6848462, "Busingen Ger", "Europe/Busingen"),
    (46.2085101, 6.1246227, "Genf", "Europe/Zurich"),
    (-29.391356857138753, 28.50989829115889, "Lesotho", "Africa/Maseru"),
    (39.93143377877638, 71.08546583764965, "Uzbek enclave1", "Asia/Tashkent"),
    (39.969915, 71.134060, "Uzbek enclave2", "Asia/Tashkent"),
    (39.862402, 70.568449, "Tajik enclave", "Asia/Dushanbe"),
    (35.7396116, -110.15029571, "Arizona Desert 1", "America/Denver"),
    (36.4091869, -110.7520236, "Arizona Desert 2", "America/Phoenix"),
    (36.10230848, -111.1882385, "Arizona Desert 3", "America/Phoenix"),
    # ocean:
    (37.81, -123.5, "Far off San Fran.", "Etc/GMT+8"),
    (50.26, -9.0, "Far off Cornwall", "Etc/GMT+1"),
    (50.5, 1, "English Channel1", "Etc/GMT"),
    (56.218, 19.4787, "baltic sea", "Etc/GMT-1"),
]

# the timezone_at_land function should return None for ocean timezones
TEST_LOCATIONS_AT_LAND = [
    (lat, lng, description, None if is_ocean_timezone(expected) else expected)
    for lat, lng, description, expected in TEST_LOCATIONS
]

EDGE_TEST_CASES = [
    # directly at the poles and the 180deg longitude border the zone result is ambiguous
    # the result should still be well defined
    # zone_name="" is interpreted as "don't care"
    (-180.0, 90.0, ""),
    (-180.0, 0.0, ""),
    (-180.0, -90.0, ""),
    (180.0, 90.0, ""),
    (180.0, 0.0, ""),
    (180.0, -90.0, ""),
    (179.999, 0.0, "Etc/GMT-12"),
    (-179.999, 0.0, "Etc/GMT+12"),
]



#!/usr/bin/env python
"""
Test script to verify that the resource management improvements work correctly.
"""

from timezonefinder import TimezoneFinder


def test_context_manager_usage():
    """Test that context managers work properly for resource management."""
    print("Testing context manager functionality...")

    # Test TimezoneFinder basic functionality
    with TimezoneFinder() as tf:
        result = tf.timezone_at(lng=-74.0059, lat=40.7128)  # New York
        print(f"New York timezone: {result}")
        assert result == "America/New_York"

    print("✓ TimezoneFinder context manager works")


def test_resource_cleanup_after_exception():
    """Test that resources are properly cleaned up even after exceptions."""
    print("Testing resource cleanup after exceptions...")

    try:
        with TimezoneFinder() as tf:
            # Test normal operation first
            result = tf.timezone_at(lng=0, lat=0)
            print(f"Timezone at (0,0): {result}")

            # Now raise an exception
            raise ValueError("Test exception")
    except ValueError as e:
        print(f"Caught expected exception: {e}")

    print("✓ Resources cleaned up properly after exception")



"""
This module provides global functions that use a singleton instance of TimezoneFinder.

Note on thread safety: These global functions are not thread-safe. If you need to use
TimezoneFinder in a multi-threaded environment, create separate TimezoneFinder instances
for each thread.
"""

from typing import List, Optional, Union

from timezonefinder.timezonefinder import TimezoneFinder
from timezonefinder.configs import CoordPairs, CoordLists

# Use a global variable to store the singleton instance
TF_INSTANCE: TimezoneFinder


def _get_tf_instance() -> TimezoneFinder:
    """Get or create the global TimezoneFinder instance

    Lazy initialization: delayed memory allocation until actually needed
    required because, the package might be used with a user defined instance
    and duplicate initialisation overhead must be avoided!
    """
    global TF_INSTANCE
    try:
        return TF_INSTANCE
    except NameError:
        # If TF_INSTANCE is not defined, create it
        TF_INSTANCE = TimezoneFinder()
    return TF_INSTANCE


def timezone_at(*, lng: float, lat: float) -> Optional[str]:
    """
    Looks up in which timezone the given coordinate is included in.
    Uses the global TimezoneFinder instance.

    Note: This function is not thread-safe. For multi-threaded environments,
    create separate TimezoneFinder instances.

    :param lng: longitude of the point in degree (-180.0 to 180.0)
    :param lat: latitude in degree (90.0 to -90.0)
    :return: the timezone name of a matching polygon or None
    """
    return _get_tf_instance().timezone_at(lng=lng, lat=lat)


def timezone_at_land(*, lng: float, lat: float) -> Optional[str]:
    """
    Computes in which land timezone a point is included in.
    Uses the global TimezoneFinder instance.

    Note: This function is not thread-safe. For multi-threaded environments,
    create separate TimezoneFinder instances.

    :param lng: longitude of the point in degree (-180.0 to 180.0)
    :param lat: latitude in degree (90.0 to -90.0)
    :return: the timezone name of a matching polygon or
        ``None`` when an ocean timezone ("Etc/GMT+-XX") has been matched.
    """
    return _get_tf_instance().timezone_at_land(lng=lng, lat=lat)


def unique_timezone_at(*, lng: float, lat: float) -> Optional[str]:
    """
    Returns the name of a unique zone within the corresponding shortcut.
    Uses the global TimezoneFinder instance.

    Note: This function is not thread-safe. For multi-threaded environments,
    create separate TimezoneFinder instances.

    :param lng: longitude of the point in degree (-180.0 to 180.0)
    :param lat: latitude in degree (90.0 to -90.0)
    :return: the timezone name of the unique zone or ``None`` if there are no or multiple zones in this shortcut
    """
    return _get_tf_instance().unique_timezone_at(lng=lng, lat=lat)


def certain_timezone_at(*, lng: float, lat: float) -> Optional[str]:
    """
    Checks in which timezone polygon the point is certainly included in.
    Uses the global TimezoneFinder instance.

    Note: This function is not thread-safe. For multi-threaded environments,
    create separate TimezoneFinder instances.

    .. note:: this is only meaningful when you have compiled your own timezone data
        where there are areas without timezone polygon coverage.
        Otherwise, some timezone will always be matched and the functionality is equal to using `.timezone_at()`
        -> useless to actually test all polygons.

    .. note:: using this function is less performant than `.timezone_at()`

    :param lng: longitude of the point in degree
    :param lat: latitude in degree
    :return: the timezone name of the polygon the point is included in or `None`
    """
    return _get_tf_instance().certain_timezone_at(lng=lng, lat=lat)


def get_geometry(
    tz_name: Optional[str] = "",
    tz_id: Optional[int] = 0,
    use_id: bool = False,
    coords_as_pairs: bool = False,
) -> List[List[Union[CoordPairs, CoordLists]]]:
    """
    Retrieves the geometry of a timezone polygon.
    Uses the global TimezoneFinder instance.

    Note: This function is not thread-safe. For multi-threaded environments,
    create separate TimezoneFinder instances.

    :param tz_name: one of the names in ``timezone_names.json`` or ``self.timezone_names``
    :param tz_id: the id of the timezone (=index in ``self.timezone_names``)
    :param use_id: if ``True`` uses ``tz_id`` instead of ``tz_name``
    :param coords_as_pairs: determines the structure of the polygon representation
    :return: a data structure representing the multipolygon of this timezone
        output format: ``[ [polygon1, hole1, hole2...], [polygon2, ...], ...]``
        and each polygon and hole is itself formatted like: ``([longitudes], [latitudes])``
        or ``[(lng1,lat1), (lng2,lat2),...]`` if ``coords_as_pairs=True``.
    """
    return _get_tf_instance().get_geometry(
        tz_name=tz_name, tz_id=tz_id, use_id=use_id, coords_as_pairs=coords_as_pairs
    )



"""
test the actually stored shortcut binary file
"""

import h3.api.numpy_int as h3
import numpy as np
import pytest


from scripts.hex_utils import surrounds_north_pole, surrounds_south_pole
from scripts.shortcuts import check_shortcut_sorting, has_coherent_sequences
from timezonefinder.configs import SHORTCUT_H3_RES
from timezonefinder.flatbuf.io.hybrid_shortcuts import (
    get_hybrid_shortcut_file_path,
    read_hybrid_shortcuts_binary,
)
from timezonefinder.timezonefinder import TimezoneFinder
from timezonefinder.utils_numba import int2coord
from timezonefinder.np_binary_helpers import (
    get_zone_ids_path,
    read_per_polygon_vector,
)
from timezonefinder.configs import DEFAULT_DATA_DIR

# Load zone IDs to determine the correct hybrid shortcut file
zone_ids_path = get_zone_ids_path(DEFAULT_DATA_DIR)
zone_ids_temp = read_per_polygon_vector(zone_ids_path)
zone_id_dtype = zone_ids_temp.dtype

hybrid_shortcut_file_path = get_hybrid_shortcut_file_path(
    zone_id_dtype, DEFAULT_DATA_DIR
)
hybrid_shortcuts = read_hybrid_shortcuts_binary(hybrid_shortcut_file_path)

# Tests now work directly with hybrid_shortcuts format

VERBOSE_TESTING = True


def latlng_to_cell(lng: float, lat: float) -> int:
    return h3.latlng_to_cell(lat, lng, SHORTCUT_H3_RES)


def test_single_shortcut_binary_exists():
    """Test that only a single .fbs binary file for the shortcut index exists in the data folder."""
    data_dir = DEFAULT_DATA_DIR

    # Find all .fbs files that could be shortcut-related
    fbs_files = list(data_dir.glob("*shortcut*.fbs"))

    # We expect exactly one shortcut .fbs file (hybrid_shortcuts_uint8.fbs or hybrid_shortcuts_uint16.fbs)
    assert len(fbs_files) == 1, (
        f"Expected exactly 1 shortcut .fbs file in {data_dir}, "
        f"but found {len(fbs_files)}: {[f.name for f in fbs_files]}"
    )

    # Verify it's the correct hybrid shortcuts file
    shortcut_file = fbs_files[0]
    assert shortcut_file.name.startswith("hybrid_shortcuts_"), (
        f"Expected hybrid shortcuts file, but found {shortcut_file.name}"
    )

    # Verify it matches the expected file based on zone_id_dtype
    expected_file = get_hybrid_shortcut_file_path(zone_id_dtype, data_dir)
    assert shortcut_file == expected_file, (
        f"Found shortcut file {shortcut_file.name} doesn't match expected {expected_file.name}"
    )


@pytest.fixture
def tf():
    return TimezoneFinder(in_memory=True)


def test_shortcut_completeness(tf):
    """Test that all points of each polygon are included in the proper shortcuts."""
    # Get access to the timezone polygons
    polygons = [tf.boundaries.coords_of(i) for i in range(tf.nr_of_polygons)]

    errors = []
    for poly_id, poly in enumerate(polygons):
        if VERBOSE_TESTING and poly_id % 100 == 0:
            print(f"\rvalidating polygon {poly_id}", end="", flush=True)

        for i, pt in enumerate(poly.T):
            # ATTENTION: int to coord conversion required!
            lng = int2coord(pt[0])
            lat = int2coord(pt[1])
            hex_id = latlng_to_cell(lng, lat)
            try:
                hybrid_value = hybrid_shortcuts[hex_id]
            except KeyError:
                errors.append(
                    f"shortcut mapping is incomplete at point ({lng}, {lat}) "
                    f"(hexagon cell id {hex_id} missing in mapping)"
                )
                continue

            # For hybrid shortcuts, check if polygon is covered
            polygon_covered = False
            if isinstance(hybrid_value, int):
                # Zone ID - check if polygon belongs to this zone
                polygon_zone = tf.zone_id_of(poly_id)
                polygon_covered = polygon_zone == hybrid_value
            else:
                # Polygon array - check if polygon is in the list
                polygon_covered = poly_id in hybrid_value

            if not polygon_covered:
                errors.append(
                    f"point #{i} ({lng}, {lat}) of polygon {poly_id} (zone {tf.zone_id_of(poly_id)}) "
                    f"is not covered by hybrid shortcut entry {hybrid_value} of cell {hex_id}"
                )

    assert not errors, f"Shortcut completeness errors: {errors[:5]}"


def test_shortcut_resolution():
    """Test that all shortcuts have the correct H3 resolution."""
    invalid_resolutions = []
    for hex_id in hybrid_shortcuts.keys():
        res = h3.get_resolution(hex_id)
        if res != SHORTCUT_H3_RES:
            invalid_resolutions.append(
                f"Hexagon {hex_id} has resolution {res}, expected {SHORTCUT_H3_RES}"
            )

    assert not invalid_resolutions, f"Resolution errors: {invalid_resolutions[:5]}"


def test_unused_polygons(tf):
    """Test that all polygons are used in at least one shortcut."""
    # Get the total number of polygons
    nr_of_polygons = tf.nr_of_polygons

    # check if all polygons are used in the shortcuts (hybrid format)
    used_polygons = set()
    for hybrid_value in hybrid_shortcuts.values():
        if isinstance(hybrid_value, int):
            # Zone ID - find all polygons belonging to this zone
            for poly_id in range(nr_of_polygons):
                if tf.zone_id_of(poly_id) == hybrid_value:
                    used_polygons.add(poly_id)
        else:
            # Polygon array
            used_polygons.update(hybrid_value)

    all_polygon_ids = set(range(nr_of_polygons))
    unused_poly_ids = all_polygon_ids - used_polygons

    assert len(unused_poly_ids) == 0, (
        f"There are {len(unused_poly_ids)} unused polygons: {unused_poly_ids}"
    )


def test_empty_shortcut():
    """Test that no shortcut entries are empty (all should have polygons or zones)."""
    # since using timezone data with ocean coverage all the cells should have polygons or zones in them
    empty_shortcuts = []
    for hex_id, hybrid_value in hybrid_shortcuts.items():
        is_empty = False
        if isinstance(hybrid_value, int):
            # Zone ID - not empty
            is_empty = False
        else:
            # Polygon array - check if empty
            is_empty = len(hybrid_value) == 0

        if is_empty:
            boundary = h3.cell_to_boundary(hex_id)[0]
            empty_shortcuts.append(f"Hexagon {hex_id} at {boundary}")

    assert not empty_shortcuts, f"Found empty shortcut entries: {empty_shortcuts[:5]}"


def test_unique_pole_cells():
    """Test that exactly one cell surrounds each pole."""
    s_pole_cells = []
    n_pole_cells = []

    for hex_id in hybrid_shortcuts.keys():
        # Check if this hex cell surrounds the poles using extracted functions
        if surrounds_south_pole(hex_id):
            s_pole_cells.append(hex_id)
        if surrounds_north_pole(hex_id):
            n_pole_cells.append(hex_id)

    assert len(s_pole_cells) == 1, (
        f"{len(s_pole_cells)} cells surround the south pole: {s_pole_cells}"
    )
    assert len(n_pole_cells) == 1, (
        f"{len(n_pole_cells)} cells surround the north pole: {n_pole_cells}"
    )


def test_shortcut_uniqueness():
    """Test that shortcuts are unique (no duplicates in polygon IDs)."""
    duplicates = []
    for hex_id, hybrid_value in hybrid_shortcuts.items():
        if isinstance(hybrid_value, int):
            # Zone ID - no duplicates by definition
            continue
        else:
            # Polygon array - check for duplicates
            polygon_ids = hybrid_value
            if len(np.unique(polygon_ids)) != len(polygon_ids):
                duplicates.append(
                    f"Shortcut {hex_id} contains duplicate polygon IDs: {polygon_ids}"
                )

    assert not duplicates, f"Shortcut uniqueness errors: {duplicates[:5]}"


def test_unique_shortcut_consistency(tf):
    """Ensure the unique shortcut entries are consistent with zone assignments."""

    # Count unique shortcuts (zone IDs in hybrid format)
    unique_shortcut_count = sum(
        1 for value in hybrid_shortcuts.values() if isinstance(value, int)
    )
    assert unique_shortcut_count > 0

    for hex_id, hybrid_value in hybrid_shortcuts.items():
        if isinstance(hybrid_value, int):
            # This is a zone ID - verify it's consistent with timezone_at
            zone_id = hybrid_value
            # Get a point in this hex cell to test
            boundary = h3.cell_to_boundary(hex_id)
            lat, lng = boundary[0]  # Get first boundary point

            # Check using TimezoneFinder's timezone_at method
            found_timezone = tf.timezone_at(lng=lng, lat=lat)
            if found_timezone is not None:
                # Convert timezone name to zone ID to compare
                expected_zone_id = tf.timezone_names.index(found_timezone)
                assert expected_zone_id == zone_id, (
                    f"Hybrid shortcut has zone {zone_id} for hex {hex_id}, "
                    f"but TimezoneFinder returns timezone '{found_timezone}' (zone {expected_zone_id})"
                )


@pytest.mark.parametrize(
    "lst,expected",
    [
        ([], True),
        ([1], True),
        ([1, 1], True),
        ([2, 3], True),
        ([2, 3, 3, 0, 0, 4], True),
        ([2, 3, 2], False),
        ([2, 3, 2, 3], False),
    ],
)
def test_has_coherent_check_fct(lst, expected):
    assert has_coherent_sequences(lst) == expected


def test_shortcut_sorting(tf):
    """Test that shortcuts are correctly sorted by zone ID and polygon size."""
    invalid_sortings = []
    for hex_id, hybrid_value in hybrid_shortcuts.items():
        if isinstance(hybrid_value, int):
            # Zone ID - no sorting to check
            continue
        else:
            # Polygon array - check sorting
            polygon_ids = hybrid_value
            try:
                check_shortcut_sorting(polygon_ids, tf.zone_ids)
            except AssertionError as e:
                invalid_sortings.append(
                    f"Invalid sorting for hexagon {hex_id}: {str(e)}"
                )

    assert not invalid_sortings, f"Shortcut sorting errors: {invalid_sortings[:5]}"



from setuptools import setup
import os

# Check whether to indicate abi3 support in a wheel name
# CFFI and setuptools use their own rules to determine ABI3 support
_abi3 = bool(os.getenv("BUILD_ABI3", ""))
print("Using ABI3 wheel suffix:", _abi3)

setup(
    cffi_modules=["timezonefinder/build.py:ffibuilder"],
    options={"bdist_wheel": {"py_limited_api": "cp39"} if _abi3 else {}},
)



from unittest import TestCase, main

from timezonefinder.timezonefinder import TimezoneFinder

tf = TimezoneFinder()


class TestCompatibility(TestCase):
    def test_with_pytz(self):
        try:
            from pytz import timezone
        except Exception:
            return
        for tz in tf.timezone_names:
            timezone(tz)


if __name__ == "__main___":
    main()



"""Utilities for working with optimized hybrid shortcut FlatBuffer data."""

from pathlib import Path
from typing import Any, Callable, Dict, List, Union
from dataclasses import dataclass

import flatbuffers
import numpy as np

from timezonefinder.configs import DEFAULT_DATA_DIR

# Static imports for uint8 schema
from timezonefinder.flatbuf.generated.shortcuts_uint8.HybridShortcutCollection import (
    HybridShortcutCollection as HybridShortcutCollectionUint8,
    HybridShortcutCollectionAddEntries as HybridShortcutCollectionAddEntriesUint8,
    HybridShortcutCollectionEnd as HybridShortcutCollectionEndUint8,
    HybridShortcutCollectionStart as HybridShortcutCollectionStartUint8,
    HybridShortcutCollectionStartEntriesVector as HybridShortcutCollectionStartEntriesVectorUint8,
)
from timezonefinder.flatbuf.generated.shortcuts_uint8.HybridShortcutEntry import (
    HybridShortcutEntryAddHexId as HybridShortcutEntryAddHexIdUint8,
    HybridShortcutEntryAddValue as HybridShortcutEntryAddValueUint8,
    HybridShortcutEntryAddValueType as HybridShortcutEntryAddValueTypeUint8,
    HybridShortcutEntryEnd as HybridShortcutEntryEndUint8,
    HybridShortcutEntryStart as HybridShortcutEntryStartUint8,
)
from timezonefinder.flatbuf.generated.shortcuts_uint8.UniqueZone import (
    UniqueZone as UniqueZoneUint8,
    UniqueZoneAddZoneId as UniqueZoneAddZoneIdUint8,
    UniqueZoneEnd as UniqueZoneEndUint8,
    UniqueZoneStart as UniqueZoneStartUint8,
)
from timezonefinder.flatbuf.generated.shortcuts_uint8.PolygonList import (
    PolygonList as PolygonListUint8,
    PolygonListAddPolyIds as PolygonListAddPolyIdsUint8,
    PolygonListEnd as PolygonListEndUint8,
    PolygonListStart as PolygonListStartUint8,
    PolygonListStartPolyIdsVector as PolygonListStartPolyIdsVectorUint8,
)
from timezonefinder.flatbuf.generated.shortcuts_uint8.ShortcutValue import (
    ShortcutValue as ShortcutValueUint8,
)

# Static imports for uint16 schema
from timezonefinder.flatbuf.generated.shortcuts_uint16.HybridShortcutCollection import (
    HybridShortcutCollection as HybridShortcutCollectionUint16,
    HybridShortcutCollectionAddEntries as HybridShortcutCollectionAddEntriesUint16,
    HybridShortcutCollectionEnd as HybridShortcutCollectionEndUint16,
    HybridShortcutCollectionStart as HybridShortcutCollectionStartUint16,
    HybridShortcutCollectionStartEntriesVector as HybridShortcutCollectionStartEntriesVectorUint16,
)
from timezonefinder.flatbuf.generated.shortcuts_uint16.HybridShortcutEntry import (
    HybridShortcutEntryAddHexId as HybridShortcutEntryAddHexIdUint16,
    HybridShortcutEntryAddValue as HybridShortcutEntryAddValueUint16,
    HybridShortcutEntryAddValueType as HybridShortcutEntryAddValueTypeUint16,
    HybridShortcutEntryEnd as HybridShortcutEntryEndUint16,
    HybridShortcutEntryStart as HybridShortcutEntryStartUint16,
)
from timezonefinder.flatbuf.generated.shortcuts_uint16.UniqueZone import (
    UniqueZone as UniqueZoneUint16,
    UniqueZoneAddZoneId as UniqueZoneAddZoneIdUint16,
    UniqueZoneEnd as UniqueZoneEndUint16,
    UniqueZoneStart as UniqueZoneStartUint16,
)
from timezonefinder.flatbuf.generated.shortcuts_uint16.PolygonList import (
    PolygonList as PolygonListUint16,
    PolygonListAddPolyIds as PolygonListAddPolyIdsUint16,
    PolygonListEnd as PolygonListEndUint16,
    PolygonListStart as PolygonListStartUint16,
    PolygonListStartPolyIdsVector as PolygonListStartPolyIdsVectorUint16,
)
from timezonefinder.flatbuf.generated.shortcuts_uint16.ShortcutValue import (
    ShortcutValue as ShortcutValueUint16,
)


@dataclass
class SchemaImports:
    """Container for schema-specific imports to eliminate magic strings."""

    # Collection functions
    collection_start: Callable[..., Any]
    collection_add_entries: Callable[..., Any]
    collection_end: Callable[..., Any]
    collection_start_entries_vector: Callable[..., Any]

    # Entry functions
    entry_start: Callable[..., Any]
    entry_add_hex_id: Callable[..., Any]
    entry_add_value_type: Callable[..., Any]
    entry_add_value: Callable[..., Any]
    entry_end: Callable[..., Any]

    # UniqueZone functions
    unique_zone_start: Callable[..., Any]
    unique_zone_add_zone_id: Callable[..., Any]
    unique_zone_end: Callable[..., Any]

    # PolygonList functions
    polygon_list_start: Callable[..., Any]
    polygon_list_add_poly_ids: Callable[..., Any]
    polygon_list_end: Callable[..., Any]
    polygon_list_start_poly_ids_vector: Callable[..., Any]

    # ShortcutValue enum
    shortcut_value: Any

    # Validation parameters
    max_zone_id: int
    dtype_name: str


@dataclass
class ReadSchemaImports:
    """Container for read-specific schema imports."""

    collection: Any
    unique_zone: Any
    polygon_list: Any
    shortcut_value: Any


def get_hybrid_shortcut_file_path(
    zone_id_dtype: np.dtype, output_path: Path = DEFAULT_DATA_DIR
) -> Path:
    """Return the path to the appropriate hybrid shortcut FlatBuffer binary file."""
    if zone_id_dtype.itemsize == 1:
        return output_path / "hybrid_shortcuts_uint8.fbs"
    elif zone_id_dtype.itemsize == 2:
        return output_path / "hybrid_shortcuts_uint16.fbs"
    else:
        raise ValueError(
            f"Unsupported zone_id_dtype: {zone_id_dtype}. Use uint8 or uint16."
        )


def _validate_zone_id_dtype(zone_id_dtype: np.dtype) -> np.dtype:
    """Validate and normalize zone ID dtype."""
    dtype = np.dtype(zone_id_dtype)
    if dtype.kind != "u":
        raise ValueError(f"Zone id dtype must be unsigned integer, got {dtype}")
    if dtype.itemsize not in (1, 2):
        raise ValueError(
            f"Zone id dtype must be 1 or 2 bytes, got {dtype.itemsize} bytes"
        )
    return dtype.newbyteorder("<")


def write_hybrid_shortcuts_flatbuffers(
    hybrid_mapping: Dict[int, Union[int, List[int]]],
    zone_id_dtype: np.dtype,
    output_file: Path,
) -> None:
    """
    Write hybrid shortcut mapping to the appropriate optimized FlatBuffer binary file.

    Args:
        hybrid_mapping: Dictionary mapping H3 hexagon IDs to either:
                       - int: unique zone ID (when all polygons share same zone)
                       - List[int]: list of polygon IDs (when multiple zones)
        zone_id_dtype: numpy dtype for zone IDs (uint8 or uint16)
        output_file: Path to save the FlatBuffer file
    """
    print(f"Writing {len(hybrid_mapping)} optimized hybrid shortcuts to {output_file}")

    dtype = _validate_zone_id_dtype(zone_id_dtype)
    _write_hybrid_shortcuts_generic(hybrid_mapping, dtype, output_file)


def _write_hybrid_shortcuts_generic(
    hybrid_mapping: Dict[int, Union[int, List[int]]],
    zone_id_dtype: np.dtype,
    output_file: Path,
) -> None:
    """Write hybrid shortcuts using the appropriate schema based on dtype."""
    if zone_id_dtype.itemsize == 1:
        # uint8 schema imports
        schema = SchemaImports(
            collection_start=HybridShortcutCollectionStartUint8,
            collection_add_entries=HybridShortcutCollectionAddEntriesUint8,
            collection_end=HybridShortcutCollectionEndUint8,
            collection_start_entries_vector=HybridShortcutCollectionStartEntriesVectorUint8,
            entry_start=HybridShortcutEntryStartUint8,
            entry_add_hex_id=HybridShortcutEntryAddHexIdUint8,
            entry_add_value_type=HybridShortcutEntryAddValueTypeUint8,
            entry_add_value=HybridShortcutEntryAddValueUint8,
            entry_end=HybridShortcutEntryEndUint8,
            unique_zone_start=UniqueZoneStartUint8,
            unique_zone_add_zone_id=UniqueZoneAddZoneIdUint8,
            unique_zone_end=UniqueZoneEndUint8,
            polygon_list_start=PolygonListStartUint8,
            polygon_list_add_poly_ids=PolygonListAddPolyIdsUint8,
            polygon_list_end=PolygonListEndUint8,
            polygon_list_start_poly_ids_vector=PolygonListStartPolyIdsVectorUint8,
            shortcut_value=ShortcutValueUint8,
            max_zone_id=255,
            dtype_name="uint8",
        )
    else:
        # uint16 schema imports
        schema = SchemaImports(
            collection_start=HybridShortcutCollectionStartUint16,
            collection_add_entries=HybridShortcutCollectionAddEntriesUint16,
            collection_end=HybridShortcutCollectionEndUint16,
            collection_start_entries_vector=HybridShortcutCollectionStartEntriesVectorUint16,
            entry_start=HybridShortcutEntryStartUint16,
            entry_add_hex_id=HybridShortcutEntryAddHexIdUint16,
            entry_add_value_type=HybridShortcutEntryAddValueTypeUint16,
            entry_add_value=HybridShortcutEntryAddValueUint16,
            entry_end=HybridShortcutEntryEndUint16,
            unique_zone_start=UniqueZoneStartUint16,
            unique_zone_add_zone_id=UniqueZoneAddZoneIdUint16,
            unique_zone_end=UniqueZoneEndUint16,
            polygon_list_start=PolygonListStartUint16,
            polygon_list_add_poly_ids=PolygonListAddPolyIdsUint16,
            polygon_list_end=PolygonListEndUint16,
            polygon_list_start_poly_ids_vector=PolygonListStartPolyIdsVectorUint16,
            shortcut_value=ShortcutValueUint16,
            max_zone_id=65535,
            dtype_name="uint16",
        )

    _write_hybrid_shortcuts_with_schema(hybrid_mapping, output_file, schema)


def _write_hybrid_shortcuts_with_schema(
    hybrid_mapping: Dict[int, Union[int, List[int]]],
    output_file: Path,
    schema: SchemaImports,
) -> None:
    """Write hybrid shortcuts using the provided schema imports."""
    builder = flatbuffers.Builder(0)
    entry_offsets = []

    # Validate zone IDs fit in dtype
    for value in hybrid_mapping.values():
        if isinstance(value, int) and value > schema.max_zone_id:
            raise ValueError(
                f"Zone ID {value} exceeds {schema.dtype_name} maximum ({schema.max_zone_id})"
            )

    for hex_id, value in hybrid_mapping.items():
        if isinstance(value, int):
            # Create UniqueZone with direct storage
            schema.unique_zone_start(builder)
            schema.unique_zone_add_zone_id(builder, value)
            unique_zone_offset = schema.unique_zone_end(builder)

            # Create entry with UniqueZone
            schema.entry_start(builder)
            schema.entry_add_hex_id(builder, hex_id)
            schema.entry_add_value_type(builder, schema.shortcut_value.UniqueZone)
            schema.entry_add_value(builder, unique_zone_offset)
            entry_offset = schema.entry_end(builder)

        else:
            # Create PolygonList
            poly_ids = list(value)
            schema.polygon_list_start_poly_ids_vector(builder, len(poly_ids))
            for i in range(len(poly_ids) - 1, -1, -1):
                builder.PrependUint16(poly_ids[i])
            poly_ids_vector = builder.EndVector()

            schema.polygon_list_start(builder)
            schema.polygon_list_add_poly_ids(builder, poly_ids_vector)
            polygon_list_offset = schema.polygon_list_end(builder)

            # Create entry with PolygonList
            schema.entry_start(builder)
            schema.entry_add_hex_id(builder, hex_id)
            schema.entry_add_value_type(builder, schema.shortcut_value.PolygonList)
            schema.entry_add_value(builder, polygon_list_offset)
            entry_offset = schema.entry_end(builder)

        entry_offsets.append(entry_offset)

    # Create entries vector
    schema.collection_start_entries_vector(builder, len(entry_offsets))
    for offset in reversed(entry_offsets):
        builder.PrependUOffsetTRelative(offset)
    entries_vector = builder.EndVector()

    # Create HybridShortcutCollection
    schema.collection_start(builder)
    schema.collection_add_entries(builder, entries_vector)
    collection = schema.collection_end(builder)

    builder.Finish(collection)

    # Write to file
    with open(output_file, "wb") as f:
        f.write(builder.Output())


def read_hybrid_shortcuts_binary(
    file_path: Path,
) -> Dict[int, Union[int, np.ndarray]]:
    """
    Read hybrid shortcut mapping from an optimized FlatBuffer binary file.

    Auto-detects whether the file uses uint8 or uint16 schema based on filename.

    Args:
        file_path: Path to the hybrid shortcut FlatBuffer file

    Returns:
        Dictionary mapping H3 hexagon IDs to either:
        - int: unique zone ID (when all polygons share same zone)
        - np.ndarray: array of polygon IDs (when multiple zones)
    """
    # Determine schema type from filename and select appropriate imports
    if "uint8" in file_path.name:
        schema = ReadSchemaImports(
            collection=HybridShortcutCollectionUint8,
            unique_zone=UniqueZoneUint8,
            polygon_list=PolygonListUint8,
            shortcut_value=ShortcutValueUint8,
        )
    elif "uint16" in file_path.name:
        schema = ReadSchemaImports(
            collection=HybridShortcutCollectionUint16,
            unique_zone=UniqueZoneUint16,
            polygon_list=PolygonListUint16,
            shortcut_value=ShortcutValueUint16,
        )
    else:
        raise ValueError(
            f"Cannot determine schema from filename: {file_path.name}. "
            "Filename must include 'uint8' or 'uint16'."
        )

    return _read_hybrid_shortcuts_with_schema(file_path, schema)


def _read_hybrid_shortcuts_with_schema(
    file_path: Path, schema: ReadSchemaImports
) -> Dict[int, Union[int, np.ndarray]]:
    """Read hybrid shortcuts using the provided schema imports."""
    with open(file_path, "rb") as f:
        buf = f.read()

    # mypy: GetRootAs is a class method on FlatBuffers classes
    collection = schema.collection.GetRootAs(buf, 0)  # type: ignore

    hybrid_mapping: Dict[int, Union[int, np.ndarray]] = {}
    for i in range(collection.EntriesLength()):
        entry = collection.Entries(i)
        hex_id = entry.HexId()

        # Determine value type and extract data
        value_type = entry.ValueType()
        value = entry.Value()

        if value_type == schema.shortcut_value.UniqueZone:
            unique_zone = schema.unique_zone()  # type: ignore
            unique_zone.Init(value.Bytes, value.Pos)
            zone_id = unique_zone.ZoneId()  # Direct zone ID, no lookup needed
            hybrid_mapping[hex_id] = int(zone_id)

        elif value_type == schema.shortcut_value.PolygonList:
            polygon_list = schema.polygon_list()  # type: ignore
            polygon_list.Init(value.Bytes, value.Pos)
            poly_ids = polygon_list.PolyIdsAsNumpy()
            hybrid_mapping[hex_id] = poly_ids

        else:
            raise ValueError(f"Unknown ShortcutValue type: {value_type}")

    return hybrid_mapping



from ast import List
from collections.abc import Iterable
import fnmatch
import os
from pathlib import Path
import random
from re import Pattern
import re
import shutil
import subprocess
import sys
import timeit
from math import log10
from typing import Callable, Iterator, Tuple, Union

import numpy as np

from scripts.utils import validate_coord_array_shape
from tests.locations import REDUCED_TIMEZONE_MAPPING
from timezonefinder import utils
from timezonefinder.configs import (
    MAX_LAT_VAL,
    MAX_LAT_VAL_INT,
    MAX_LNG_VAL,
    MAX_LNG_VAL_INT,
    PACKAGE_DIR,
)
from timezonefinder.polygon_array import PolygonArray
from timezonefinder.utils_numba import convert2coords


#######################
# PATH CONSTANTS
#######################

PROJECT_ROOT = PACKAGE_DIR.parent
DIST_DIR = PROJECT_ROOT / "dist"


# Command constants
# BUILD_SDIST_CMD = [sys.executable, "setup.py", "sdist"]
BUILD_SDIST_CMD = ["uv", "build", "-v", "--sdist"]
BUILD_WHEEL_CMD = [sys.executable, "setup.py", "bdist_wheel"]
# BUILD_WHEEL_CMD = ["uv", "build", "-v", "--wheel"]


# for reading coordinates
boundaries_dir = utils.get_boundaries_dir()
boundaries = PolygonArray(data_location=boundaries_dir, in_memory=True)

#######################
# UTILITY FUNCTIONS
#######################


def run_command(
    cmd: list, capture_output: bool = False, cwd: Path = PROJECT_ROOT
) -> subprocess.CompletedProcess:
    """Run a command and handle errors appropriately."""
    print(f"Running command: {' '.join(cmd)}")
    try:
        return subprocess.run(
            cmd,
            check=True,
            capture_output=capture_output,
            text=capture_output,
            cwd=str(cwd),
        )
    except subprocess.CalledProcessError as e:
        # Include the stdout/stderr in the error message if available
        error_msg = str(e)
        if capture_output and e.stdout:
            error_msg += f"\nStdout: {e.stdout}"
        if capture_output and e.stderr:
            error_msg += f"\nStderr: {e.stderr}"
        raise subprocess.CalledProcessError(
            e.returncode, e.cmd, e.output, e.stderr
        ) from None


def build_wheel() -> Path:
    """Build wheel distribution and return its path."""
    # TODO reuse with DistributionFilesFixture (found in test_package_contents.py)
    # Clean up dist directory if it exists
    if DIST_DIR.exists():
        shutil.rmtree(DIST_DIR)

    # Build the wheel
    run_command(BUILD_WHEEL_CMD, cwd=str(PROJECT_ROOT))

    # Find the wheel file
    wheels = list(DIST_DIR.glob("*.whl"))
    assert wheels, "No wheel file found in dist/"
    return wheels[0]


def build_sdist() -> Path:
    """Build the distribution using the configured build command and return the path to the archive."""
    # Clean up dist directory if it exists
    if DIST_DIR.exists():
        shutil.rmtree(DIST_DIR)

    run_command(BUILD_SDIST_CMD, cwd=str(PROJECT_ROOT))

    dist_files = file_path_iterator(DIST_DIR, relative=False)
    # Find the generated .tar.gz file in the dist directory
    sdist_files = list(filter_paths(dist_files, "*.tar.gz"))
    assert len(sdist_files) == 1, "Expected exactly one .tar.gz distribution file"
    sdist = sdist_files[0]
    print(f"Found distribution file: {sdist}")
    return sdist


def file_path_iterator(
    path: Path = PROJECT_ROOT, relative: bool = False
) -> Iterator[Path]:
    """
    Recursively iterate over all files in the given path.

    Args:
        path: The root path to start the iteration from (default: PROJECT_ROOT)

    Yields:
        Path objects for each file found
    """
    assert isinstance(path, Path), "path must be a Path object"
    # assert path.is_dir(), f"path must be a directory, got {path}"

    # recursively walk through the directory
    for root, _, files in os.walk(path):
        for file in files:
            # yield the full path to the file
            # using Path to ensure compatibility with different OS path formats
            file_path = Path(root) / file
            if relative:
                # yield relative to the project root
                file_path = file_path.relative_to(path)
            yield file_path


def matches_pattern(path: Path, pattern: Union[str, Pattern, None]) -> bool:
    r"""
    Check if a path matches a given pattern.

    Args:
        path: The path to check
        pattern: A glob pattern string or compiled regex pattern to match against
                 If None, always returns True (matches everything)
                 you can use:
                   - Simple filename patterns: '*.py' matches any Python file
                   - Directory patterns: 'tests/*.py' matches Python files in tests directory
                   - Path patterns: '*/data/*.json' matches JSON files in any data directory

    Returns:
        bool: True if the path matches the pattern, False otherwise

    Examples:
        # Check if file matches a glob pattern (filename only)
        is_python_file = matches_pattern(Path('script.py'), '*.py')  # True

        # Match against full path including directories
        in_tests_dir = matches_pattern(Path('tests/test_data.py'), 'tests/*.py')  # True

        # Match files in any data directory
        data_file = matches_pattern(Path('src/data/config.json'), '*/data/*.json')  # True

        # Check with regex pattern against full path
        import re
        is_test_file = matches_pattern(
            Path('tests/unit/test_utils.py'),
            re.compile(r'tests/.*\.py$')
        )  # True

        # Always matches when pattern is None
        matches_all = matches_pattern(Path('any_file.txt'), None)  # True
    """
    if pattern is None:
        return True
    assert isinstance(path, Path), "path must be a Path object"
    # Remove assert for is_file() to allow matching directories too
    assert isinstance(pattern, (str, re.Pattern)), (
        "pattern must be a string or a compiled regex pattern"
    )

    # Get the relative path as string for matching
    path_str = str(path)
    if isinstance(pattern, str):
        if pattern.endswith("/"):
            # pattern points to a directory
            # all content should be matched
            pattern = pattern + "*"

        # For string patterns, check against both the full path
        # Try matching against the full path first
        return fnmatch.fnmatch(path_str, pattern)
    elif isinstance(pattern, re.Pattern):
        # For regex patterns, always match against the full path
        return bool(pattern.search(path_str))


def filter_paths(
    paths: Iterator[Path],
    pattern: Union[str, Pattern, None] = None,
    include_matches: bool = True,
) -> Iterator[Path]:
    """
    Filter paths based on a pattern, either keeping matches or non-matches.

    Args:
        paths: An iterator of Path objects to filter (can be files or directories)
        pattern: A glob pattern string or compiled regex pattern to filter by
                 If None, behavior depends on include_matches
                 Patterns can include directory parts, e.g. 'tests/*.py'
        include_matches: If True, yield paths that match the pattern
                         If False, yield paths that don't match the pattern

    Yields:
        Path objects that match (or don't match) the pattern based on include_matches
    """
    for path in paths:
        is_match = matches_pattern(path, pattern)
        if (
            is_match == include_matches
        ):  # Yield when match status matches desired include status
            yield path


def any_filter_paths(
    paths: Iterator[Path], patterns: Iterable[str], include_matches: bool = True
) -> Iterator[Path]:
    """Filter paths by multiple patterns, yielding paths that match any of the patterns."""
    for path in paths:
        is_match = any(matches_pattern(path, pattern) for pattern in patterns)
        if is_match == include_matches:
            yield path


def ocean2land(test_locations):
    for lat, lng, description, expected in test_locations:
        if utils.is_ocean_timezone(expected):
            expected = None
        yield lat, lng, description, expected


def check_geometry(geometry_obj: List):
    coords = geometry_obj[0][0]
    assert len(coords) == 2, (
        "the polygon does not consist of two latitude longitude lists"
    )
    x_coords, y_coords = coords
    nr_x_coords = len(x_coords)
    nr_y_coords = len(y_coords)
    assert nr_x_coords > 2, "a polygon must consist of more than 2 coordinates"
    assert nr_x_coords == nr_y_coords, (
        "the amount of x and y coordinates (lng, lat) must be equal"
    )


def check_pairwise_geometry(geometry_obj: List):
    # list of all coord pairs of the first polygon
    cord_pairs = geometry_obj[0][0]
    assert len(cord_pairs) > 2, "a polygon must consist of more than 2 coordinates"
    first_coord_pair = cord_pairs[0]
    assert len(first_coord_pair) == 2, (
        "the polygon does not consist of coordinate pairs as expected."
    )


def is_valid_lng_int(x: int) -> bool:
    return -MAX_LNG_VAL_INT <= x <= MAX_LNG_VAL_INT


def is_valid_lat_int(y: int) -> bool:
    return -MAX_LAT_VAL_INT <= y <= MAX_LAT_VAL_INT


def is_valid_lng_int_vec(arr) -> bool:
    return np.all((-MAX_LNG_VAL_INT <= arr) & (arr <= MAX_LNG_VAL_INT))


def is_valid_lat_int_vec(arr) -> bool:
    return np.all((-MAX_LAT_VAL_INT <= arr) & (arr <= MAX_LAT_VAL_INT))


def validate_polygon_coordinates(coords: np.ndarray):
    """Helper function to validate polygon coordinates format and values."""
    validate_coord_array_shape(coords)

    # test whether the coordinates are within valid ranges
    x_coords, y_coords = coords
    # apply to every coordinate

    assert is_valid_lng_int_vec(x_coords)
    assert is_valid_lat_int_vec(y_coords)


def timefunc(function: Callable, *args):
    def wrap():
        function(*args)

    timer = timeit.Timer(wrap)
    nr_runs = 1
    t_in_sec = timer.timeit(nr_runs)
    return t_in_sec


def proto_test_case(data, fct):
    for input, expected_output in data:
        # print(input, expected_output, fct(input))
        actual_output = fct(input)
        if actual_output != expected_output:
            print(
                "input: {} expected: {} got: {}".format(
                    input, expected_output, actual_output
                )
            )
        assert actual_output == expected_output


def time_preprocess(time):
    valid_digits = 4
    zero_digits = abs(min(0, int(log10(time))))
    digits_to_print = zero_digits + valid_digits
    return str(round(time, digits_to_print)) + "s"


def get_rnd_query_pt() -> Tuple[float, float]:
    lng = random.uniform(-MAX_LNG_VAL, MAX_LNG_VAL)
    lat = random.uniform(-MAX_LAT_VAL, MAX_LAT_VAL)
    return lng, lat


def get_rnd_poly_int() -> np.ndarray:
    max_poly_id = len(boundaries) - 1
    poly_id = random.randint(0, max_poly_id)
    poly = boundaries.coords_of(poly_id)
    return poly


def get_rnd_poly() -> np.ndarray:
    poly = get_rnd_poly_int()
    coords = convert2coords(poly)
    return np.array(coords)


def convert_inside_polygon_input(lng: float, lat: float):
    x, y = utils.coord2int(lng), utils.coord2int(lat)
    return x, y


def get_pip_test_input() -> Tuple[int, int, np.ndarray]:
    # one test polygon + one query point
    lng, lat = get_rnd_query_pt()
    x, y = convert_inside_polygon_input(lng, lat)
    poly_int = get_rnd_poly_int()
    return x, y, poly_int


def convert_to_reduced_timezone(timezone: str) -> str:
    """
    Convert a timezone to its reduced version using the provided mapping.

    Args:
        timezone: The original timezone string.
        mapping: A dictionary mapping original timezones to their reduced versions.

    Returns:
        The reduced timezone if found in the mapping, otherwise the original timezone.
    """
    return REDUCED_TIMEZONE_MAPPING.get(timezone, timezone)


def single_location_test(func, lat, lng, description, expected_orig):
    result = func(lng=lng, lat=lat)
    func_name = func.__name__
    expected = convert_to_reduced_timezone(expected_orig)
    assert result == expected, (
        f"{func_name}({lng}, {lat}) [{description}] should return {expected}, got {result}"
    )
